from __future__ import annotations

import escnn.group
from escnn.group import Group, GroupElement
from escnn.group import IrreducibleRepresentation, Representation
from escnn.group import utils

import numpy as np
import math

from typing import Tuple, Callable, Iterable, List, Dict, Any

__all__ = ["DihedralGroup"]


class DihedralGroup(Group):
    
    PARAM = 'int'
    PARAMETRIZATIONS = [
        'int',          # integer in 0, 1, ..., N-1
        'radians',      # real in 0., 2pi/N, ... i*2pi/N, ...
        # 'C',            # point in the unit circle (i.e. cos and sin of 'radians')
        'MAT',  # 2x2 rotation matrix
    ]

    def __init__(self, N: int):
        r"""
        Build an instance of the dihedral group :math:`D_N` which contains reflections and ``N`` discrete planar
        rotations.
        The order of the group :math:`D_N` is :math:`2N`.
        
        The group elements are
        :math:`\{e,\, r,\, r^2,\, r^3,\, \dots,\, r^{N-1},\, f,\, rf,\, r^2f,\, r^3f,\, \dots,\, r^{N-1}f\}`,
        so an element of this group is either a rotation :math:`r^k` or a reflection :math:`r^kf` along an axis.
        
        Any group element is either a discrete rotation :math:`r^k` by an angle :math:`k\frac{2\pi}{N}` or a
        reflection :math:`f` followed by a rotation, i.e. :math:`r^kf`.
        As in :class:`~escnn.group.CyclicGroup`, combination of rotations behaves like arithmetic modulo ``N``, so
        :math:`r^a \cdot r^b = r^{\ a + b \!\! \mod \!\! N}\ `.
        Two reflections gives the identity :math:`f \cdot f = e` and a reflection commutes with a rotation by inverting
        it, i.e. :math:`r^k \cdot f = f \cdot r^{-k}`.
        
        A group element :math:`r^kf^j` is implemented as a pair :math:`(j, k)` with :math:`j \in \{0, 1\}` and
        and :math:`k \in \{0, \dots, N-1\}`.
        
        
        
        Subgroup Structure.
        
        A subgroup of :math:`D_{2N}` is identified by a tuple ``id`` :math:`(k, M)`.
        
        Here, :math:`M` is a positive integer indicating the number of discrete rotations in the subgroup while
        :math:`k` is either ``None`` or an integer in :math:`\{0, \dots, \frac{N}{M}-1\}`. If :math:`k` is ``None``,
        the subgroup does not contain any reflections. Otherwise, the subgroup contains the reflection :math:`r^k f`
        along the axis of the current group rotated by :math:`k\frac{\pi}{N}`.
        The order :math:`M` has to divide the rotation order :math:`N` of the current group (:math:`D_{2N}`).
        
        Valid combinations are:
        
        - (``None``, :math:`M`): restrict to the cyclic subgroup :math:`C_M` generated by :math:`\langle r^{N/M} \rangle`.
        
        - (:math:`k`, :math:`M`): restrict to the dihedral subgroup :math:`D_{M}` generated by :math:`\langle r^{N/M}, r^{k}f \rangle`
        
        In particular:
        
        - (``None``, :math:`1`): restrict to the cyclic subgroup of order 1 containing only the identity
        
        - (:math:`0`, :math:`1`): restrict to the reflection group generated by :math:`\langle f \rangle`
        
        - (:math:`0`, :math:`M`): restrict to the dihedral subgroup :math:`D_{M}` generated by :math:`\langle r^{N/M}, f \rangle`
        
        - (:math:`k`, :math:`1`): restrict to the reflection group generated by :math:`\langle r^{k}f \rangle = \{e, r^{k}f\}`
        
        
        Args:
            N (int): number of discrete rotations in the group
            
        Attributes:
            
            ~.reflection: the reflection element
            ~.rotation_order (int): the number of discrete rotations in this group (equal to the parameter ``N``)
            
        """
        
        assert (isinstance(N, int) and N > 0), N
        
        super(DihedralGroup, self).__init__("D%d" % N, False, False)
        
        self.rotation_order = N
        
        self._elements = [self.element((0, i)) for i in range(N)] + [self.element((1, i)) for i in range(N)]

        self._identity = self.element((0, 0))
        self.reflection = self.element((1, 0))
        
        self._build_representations()

    @property
    def generators(self) -> List[GroupElement]:
        if self.rotation_order > 1:
            return [self.element((0, 1)), self.element((1, 0))]
        else:
            return [self.element((1, 0))]

    @property
    def identity(self) -> GroupElement:
        return self._identity
    
    @property
    def elements(self) -> List[GroupElement]:
        return self._elements
     
    @property
    def _keys(self) -> Dict[str, Any]:
        return {'N': self.rotation_order}
    
    @property
    def subgroup_trivial_id(self):
        return (None, 1)

    @property
    def subgroup_self_id(self):
        return (0, self.rotation_order)

    ###########################################################################
    # METHODS DEFINING THE GROUP LAW AND THE OPERATIONS ON THE GROUP'S ELEMENTS
    ###########################################################################

    def _inverse(self, element: Tuple[int, int], param: str = PARAM) -> Tuple[int, int]:
        r"""
        Returns the inverse element of the input element.
        Given the element :math:`r^kf^j` as a pair :math:`(j, k)`,
        the method returns :math:`r^{-k}` (as :math:`(0, -k)`) if :math:`f = 0` and
        :math:`fr^{-k}=r^kf` (as :math:`(1, k)`) otherwise.
        
        Args:
            element (tuple): a group element :math:`r^kf^j` as a pair :math:`(j, k)`

        Returns:
            its inverse
           
        """
        element = self._change_param(element, p_from=param, p_to='int')
        inverse = element[0], (-element[1] * (-1 if element[0] else 1)) % self.rotation_order
        return self._change_param(inverse, p_from='int', p_to=param)

    def _combine(self, e1: Tuple[int, int], e2: Tuple[int, int],
                 param: str = PARAM,
                 param1: str = None,
                 param2: str = None
    ) -> Tuple[int, float]:
        r"""
        Return the combination of the two input elements.
        Given two input element :math:`r^af^b` and :math:`r^cf^d`, the method returns
        :math:`r^af^b \cdot r^cf^d`.

        Args:
            e1 (tuple): a group element :math:`r^af^b` as a pair :math:`(b, a)`
            e2 (tuple): another element :math:`r^cf^d` as a pair :math:`(d, c)`

        Returns:
            their combination :math:`r^af^b \cdot r^cf^d`
           
        """
        if param1 is None:
            param1 = param
        if param2 is None:
            param2 = param

        e1 = self._change_param(e1, p_from=param1, p_to='int')
        e2 = self._change_param(e2, p_from=param2, p_to='int')
        product = (e1[0] + e2[0]) % 2, (e1[1] + (-1 if e1[0] else 1) * e2[1]) % self.rotation_order
        return self._change_param(product, p_from='int', p_to=param)

    def _is_element(self, element: Tuple[int, int], param: str = PARAM, verbose: bool = False) -> bool:
        element = self._change_param(element, p_from=param, p_to='int')
        if isinstance(element, tuple) and len(element) == 2 and isinstance(element[0], int) and isinstance(element[1],
                                                                                                           int):
            return element[0] in {0, 1} and 0 <= element[1] < self.rotation_order
        else:
            return False

    def _equal(self, e1: Tuple[int, int], e2: Tuple[int, int],
               param: str = PARAM,
               param1: str = None,
               param2: str = None
    ) -> bool:
        r"""

        Check if the two input values corresponds to the same element.

        Args:
            e1 (tuple): an element
            e2 (tuple): another element

        Returns:
            whether they are the same element

        """
        if param1 is None:
            param1 = param
        if param2 is None:
            param2 = param

        e1 = self._change_param(e1, p_from=param1, p_to='int')
        e2 = self._change_param(e2, p_from=param2, p_to='int')
        return e1[0] == e2[0] and e1[1] == e2[1]
    
    def _hash_element(self, element: Tuple[int, int], param: str = PARAM):
        element = self._change_param(element, p_from=param, p_to='int')
        element = element[0], element[1] % self.rotation_order
        return hash(element)

    def _repr_element(self, element: Tuple[int, int], param: str = PARAM):
        element = self._change_param(element, p_from=param, p_to='int')
        return "({}, {}[2pi/{}])".format('+' if not element[0] else '-', element[1], self.rotation_order)
    
    def _change_param(self, element: Tuple, p_from: str, p_to: str):
        assert p_from in self.PARAMETRIZATIONS
        assert p_to in self.PARAMETRIZATIONS
        
        assert isinstance(element, tuple) and len(element) == 2
        assert element[0] in [0, 1]
        
        flip, rotation = element

        if p_from == 'MAT':
            assert isinstance(rotation, np.ndarray)
            assert rotation.shape == (2, 2)
            assert np.isclose(np.linalg.det(rotation), 1.)
            assert np.allclose(rotation @ rotation.T, np.eye(2))
    
            cos = (rotation[0, 0] + rotation[1, 1]) / 2.
            sin = (rotation[1, 0] - rotation[1, 0]) / 2.
    
            rotation = np.arctan2(sin, cos)
            p_from = 'radians'

        # convert to INT
        if p_from == 'int':
            assert isinstance(rotation, int)
        elif p_from == 'radians':
            assert isinstance(rotation, float)
            if not utils.cycle_isclose(rotation, 0., 2 * np.pi / self.rotation_order):
                raise ValueError()
            rotation = int(round(self.rotation_order * rotation / (2 * np.pi))) % self.rotation_order
        else:
            raise ValueError('Parametrization {} not recognized'.format(p_from))

        # convert from INT
        if p_to == 'int':
            return flip, rotation
        elif p_to == 'radians':
            rotation = rotation * (2 * np.pi) / self.rotation_order
            return flip, rotation
        elif p_to == 'MAT':
            rotation = rotation * (2 * np.pi) / self.rotation_order
            cos = np.cos(rotation)
            sin = np.sin(rotation)
            rotation = np.array(([
                [cos, -sin],
                [sin, cos],
            ]))
            return flip, rotation
        else:
            raise ValueError('Parametrization {} not recognized'.format(p_to))

    ###########################################################################
    
    def sample(self) -> GroupElement:
        return self.element((np.random.randint(0, 2), np.random.randint(0, self.rotation_order)))

    def testing_elements(self) -> Iterable[GroupElement]:
        r"""
        A finite number of group elements to use for testing.
        """
        return iter(self._elements)
    
    def __eq__(self, other):
        if not isinstance(other, DihedralGroup):
            return False
        else:
            return self.name == other.name and self.rotation_order == other.rotation_order
    
    def _subgroup(self, id: Tuple[int, int]) -> Tuple[
        Group,
        Callable[[GroupElement], GroupElement],
        Callable[[GroupElement], GroupElement]
    ]:
        r"""
        Restrict the current group :math:`D_{2N}` to the subgroup identified by the input ``id``, where ``id`` is a
        tuple :math:`(k, M)`.
        
        Args:
            id (tuple): the identification of the subgroup

        Returns:
            a tuple containing

                - the subgroup,

                - a function which maps an element of the subgroup to its inclusion in the original group and

                - a function which maps an element of the original group to the corresponding element in the subgroup (returns None if the element is not contained in the subgroup)
              
        """
        
        assert isinstance(id, tuple) and len(id) == 2
        assert id[0] is None or isinstance(id[0], int)
        assert isinstance(id[1], int)
        assert id[1] >= 1

        axis = id[0]
        order = id[1]
        
        assert self.rotation_order % order == 0, \
            "Error! The rotations order of the subgroups of a dihedral group has to divide the rotations order of the overgroup." \
            " %d does not divide %d " % (order, self.rotation_order)

        assert axis is None or 0 <= axis < self.rotation_order # // order

        ratio = self.rotation_order // order
        
        if id[0] is not None and id[1] == 1:
            
            # take the elements of the group generated by "r^axis f"
            sg = escnn.group.cyclic_group(2)

            # parent_mapping = lambda e, axis=axis: self.element((e._element, axis * e._element))
            # child_mapping = lambda e, axis=axis, sg=sg: None if e._element[1] != e._element[0] * axis else sg.element(e._element[0])
            parent_mapping = flip_to_dn(axis, self)
            child_mapping = dn_to_flip(axis, sg)

        elif id[0] is None:
            # take the elements of the group generated by "r^ratio"
            sg = escnn.group.cyclic_group(order)
            # parent_mapping = lambda e, ratio=ratio: self.element((0, e._element * ratio))
            # child_mapping = lambda e, ratio=ratio, sg=sg: None if (e._element[0] != 0 or e._element[1] % ratio > 0) else sg.element(int(e._element[1] / ratio))
            parent_mapping = cn_to_dn(self)
            child_mapping = dn_to_cn(sg)

        else:
            # take the elements of the group generated by "r^ratio" and "r^axis f"
            sg = escnn.group.dihedral_group(order)
            # parent_mapping = lambda e, ratio=ratio, axis=axis: self.element((e._element[0], e._element[1] * ratio + e._element[0] * axis))
            # child_mapping = lambda e, ratio=ratio, axis=axis, sg=sg: None if (e._element[1] - e._element[0] * axis) % ratio > 0 else sg.element((e._element[0], int((e._element[1] - e._element[0] * axis) / ratio)))
            parent_mapping = dm_to_dn(axis, self)
            child_mapping = dn_to_dm(axis, sg)

        return sg, parent_mapping, child_mapping

    def _combine_subgroups(self, sg_id1, sg_id2):
    
        sg_id1 = self._process_subgroup_id(sg_id1)
        sg1, inclusion, restriction = self.subgroup(sg_id1)
        sg_id2 = sg1._process_subgroup_id(sg_id2)
        
        if sg_id1[0] is None:
            return (None, sg_id2)
        elif sg_id1[1] == 1:
            return sg_id1[0] if sg_id2 == 2 else None, 1
        elif sg_id2[0] is None:
            return sg_id2
        else:
            flip = sg_id1[0] + inclusion(sg1.element((0, sg_id2[0]))).to('int')[1]
            return (flip,) + sg_id2[1:]

    def _restrict_irrep(self, irrep: Tuple, id: Tuple[int, int]) -> Tuple[np.matrix, List[Tuple]]:
        r"""
        Restrict the input irrep of current group :math:`D_{2n}` to the subgroup identified by "id".
        More precisely, "id" is a tuple :math:`(k, m)`, where :math:`m` is a positive integer indicating the number of
        rotations in the subgroup while :math:`k` is either None (no flips in the subgroup) or an integer in
        :math:`[0, \frac{n}{m}-1]` (indicating the axis of flip in the subgroup).
        The order :math:`m` has to divide the rotation order :math:`n` of the current group (:math:`D_{2n}`).

        Valid combinations are:
        - (None, m): restrict to the cyclic subgroup with order "m" :math:`C_m` generated by :math:`\langle r^{(n/m)} \rangle`.
        - (1, m): restrict to the dihedral subgroup with order "2m" :math:`D_{2m}` generated by :math:`\langle r^{n/m}, f \rangle`
        - (k, 1): restrict to the cyclic subgroup of order 2 :math:`C_2` generated by the flip :math:`\langle r^{k}f \rangle = \{e, r^{k}f\}`
        - (None, 1): restrict to the cyclic subgroup of order 1 :math:`C_1` containing only the identity
        - (k, m): restrict to the dihedral subgroup with order "2m" :math:`D_{2m}` generated by :math:`\langle r^{n/m}, r^{k}f \rangle`
        
        Args:
            irrep (tuple): the identifier of the irrep to restrict
            id (tuple): the identification of the subgroup

        Returns:
            a pair containing the change of basis and the list of irreps of the subgroup which appear in the restricted irrep
            
        """

        irr = self.irrep(*irrep)
        
        sg, _, _ = self.subgroup(id)
        
        irreps = []
        change_of_basis = None

        if id[0] is not None and id[1] == 1:
            j = irr.attributes["flip_frequency"]
            k = irr.attributes["frequency"]
            if k == self.rotation_order/2:
                j = (j+id[0]) % 2
            change_of_basis = np.eye(irr.size)
            if irr.size > 1:
                irreps.append((0,))
                change_of_basis = utils.psi(0.5 * id[0] * 2 * np.pi / self.rotation_order, k)
            irreps.append((j,))
        
        elif id[0] is None:
            
            order = id[1]
            
            f = irr.attributes["frequency"] % order
            if f > order / 2:
                f = order - f
                change_of_basis = utils.chi(1)
            else:
                change_of_basis = np.eye(irr.size)
            
            r = (f,)
            irreps.append(r)
            if sg.irrep(*r).size < irr.size:
                irreps.append(r)
            
        elif id[0] is not None and id[1] > 1:
            
            order = id[1]
            f = irr.attributes["frequency"]
            j = irr.attributes["flip_frequency"]
            if f == self.rotation_order/2:
                j = (j+id[0]) % 2
            k = f % order
            
            if k > order / 2:
                k = order - k
                change_of_basis = utils.chi(1)
            else:
                change_of_basis = np.eye(irr.size)
                
            r = (j, k)
            
            if sg.irrep(*r).size < irr.size:
                irreps.append((0, k))
                irreps.append(r)
            else:
                irreps.append(r)
            
            if irr.size == 2:
                # change_of_basis = phi(0.5 * id[0], f) @ change_of_basis
                change_of_basis = utils.psi(0.5 * id[0] * 2 * np.pi / self.rotation_order, f) @ change_of_basis

        else:
            raise ValueError(f"id '{id}' not recognized")
        
        return change_of_basis, irreps
    
    def _build_representations(self):
        r"""
        Build the irreps and the regular representation for this group

        """
        
        n = self.rotation_order
        
        # Build all the Irreducible Representations
        
        # add Trivial representation
        j, k = 0, 0
        self.irrep(j, k)
        
        j = 1
        
        for k in range(0, int(n//2)+1):
            self.irrep(j, k)
        
        if n % 2 == 0:
            j = 0
            self.irrep(j, k)
            
        # Build all Representations
        
        # add all the irreps to the set of representations already built for this group
        self.representations.update(**{irr.name : irr for irr in self.irreps()})

        # build the regular representation
        
        # N.B.: it represents the LEFT-ACTION of the elements
        self.representations['regular'] = self.regular_representation

    def _build_quotient_representations(self):
        r"""
        Build all the quotient representations for this group

        """
        for n in range(2, int(math.ceil(math.sqrt(self.rotation_order)))):
            if self.rotation_order % n == 0:
                for f in range(2):
                    sg_id = (f, n)
                    self.quotient_representation(sg_id)
        
    @property
    def trivial_representation(self) -> Representation:
        return self.irrep(0, 0)

    def irrep(self, j: int, k: int) -> IrreducibleRepresentation:
        r"""
        Build the irrep with reflecion and rotation frequencies :math:`j` (reflection) and :math:`k` (rotation) of the
        current dihedral group. Note: the frequencies has to be non-negative integers, i.e.
        :math:`j \in [0, 1]` and :math:`k \in \{0, \dots, \left\lfloor N/2 \right\rfloor \}`,
        where :math:`N` is the rotational order of the group (the number of rotations, i.e. the half the group order).
        
        If :math:`N` is odd, valid parameters are :math:`(0, 0)`, :math:`(1, 0)`, :math:`(1, 1)` ... :math:`(1, \left\lfloor N/2 \right\rfloor)`.
        If :math:`N` is even, the group also has the irrep :math:`(0, N/2)`.
        
        Args:
            j (int): the frequency of the reflections in the irrep
            k (int): the frequency of the rotations in the irrep

        Returns:
            the corresponding irrep

        """
        
        N = self.rotation_order
        id = (j, k)
        
        assert j in [0, 1]
        assert 0 <= k <= N//2
    
        name = f"irrep_{j},{k}"

        if id not in self._irreps:
    
            irrep = _build_irrep_dn(j, k)
            character = _build_char_dn(j, k)

            if j == 0:
                
                if k == 0:
                    # Trivial representation
                    supported_nonlinearities = ['pointwise', 'norm', 'gated', 'gate', 'concatenated']
                    self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 1, 'R',
                                                                  supported_nonlinearities=supported_nonlinearities,
                                                                  # trivial=True,
                                                                  character=character,
                                                                  frequency=k,
                                                                  flip_frequency=j
                                                                  )

                elif N % 2 == 0 and k == N//2:
                    
                    supported_nonlinearities = ['norm', 'gated', 'concatenated']
                    self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 1, 'R',
                                                                  supported_nonlinearities=supported_nonlinearities,
                                                                  character=character,
                                                                  frequency=k,
                                                                  flip_frequency=j
                                                                  )
                else:
                    raise ValueError(f"Error! Flip frequency {j} and rotational frequency {k} don't correspond to any irrep of the group {self.name}!")

            else:
                
                if k == 0:
                    # Trivial on Cyclic subgroup Representation
                    supported_nonlinearities = ['norm', 'gated', 'concatenated']
                    self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 1, 'R',
                                                                  supported_nonlinearities=supported_nonlinearities,
                                                                  character=character,
                                                                  frequency=k,
                                                                  flip_frequency=j
                                                                  )

                elif N % 2 == 0 and k == N / 2:
    
                    # 1 dimensional Irreducible representation (only for groups with an even number of rotations)
                    supported_nonlinearities = ['norm', 'gated', 'concatenated']
                    self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 1, 'R',
                                                                  supported_nonlinearities=supported_nonlinearities,
                                                                  character=character,
                                                                  frequency=k,
                                                                  flip_frequency=j
                                                                  )
                else:
                    # 2 dimensional Irreducible Representations
                    supported_nonlinearities = ['norm', 'gated']
                    self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 2, 'R',
                                                                  supported_nonlinearities=supported_nonlinearities,
                                                                  character=character,
                                                                  frequency=k,
                                                                  flip_frequency=j
                                                                  )

        return self._irreps[id]

    _cached_group_instances = {}

    @classmethod
    def _generator(cls, N: int) -> 'DihedralGroup':
        if N not in cls._cached_group_instances:
            cls._cached_group_instances[N] = DihedralGroup(N)
    
        return cls._cached_group_instances[N]


def _build_irrep_dn(j: int, k: int):
    
    def irrep(element: GroupElement, j=j, k=k) -> np.ndarray:
        N = element.group.rotation_order
        
        if j == 0:
            if k == 0:
                return np.eye(1)
            elif N % 2 == 0 and k == N // 2:
                return np.array([[np.cos(k * element.to('radians')[1])]])
            else:
                raise ValueError(
                    f"Error! Flip frequency {j} and rotational frequency {k} don't correspond to any irrep of the group {element.group.name}!")
        else:
            if k == 0:
                # Trivial on Cyclic subgroup Representation
                return np.array([[-1 if element.to('int')[0] else 1]])
            elif N % 2 == 0 and k == N / 2:
                e = element.to('radians')
                # 1 dimensional Irreducible representation (only for groups with an even number of rotations)
                return np.array([[np.cos(k * e[1]) * (-1 if e[0] else 1)]])
            else:
                # 2 dimensional Irreducible Representations
                e = element.to('radians')
                return utils.psichi(e[1], e[0], k=k)
            
    return irrep


def _build_char_dn(j: int, k: int):
    
    def character(element: GroupElement, j=j, k=k) -> float:
        N = element.group.rotation_order
        
        if j == 0:
            if k == 0:
                return 1.
            elif N % 2 == 0 and k == N // 2:
                return np.cos(k * element.to('radians')[1])
            else:
                raise ValueError(
                    f"Error! Flip frequency {j} and rotational frequency {k} don't correspond to any irrep of the group {element.group.name}!")
        else:
            if k == 0:
                # Trivial on Cyclic subgroup Representation
                return -1 if element.to('int')[0] else 1
            elif N % 2 == 0 and k == N / 2:
                e = element.to('radians')
                # 1 dimensional Irreducible representation (only for groups with an even number of rotations)
                return np.cos(k * e[1]) * (-1 if e[0] else 1)
            else:
                # 2 dimensional Irreducible Representations
                e = element.to('radians')
                return 0 if e[0] else (2 * np.cos(k * e[1]))
            
    return character


# Cyclic ###############################

def dn_to_cn(cn: escnn.group.CyclicGroup):

    def _map(e: GroupElement, cn=cn):
        assert isinstance(e.group, DihedralGroup)

        flip, rotation = e.to('int')

        ratio = e.group.rotation_order // cn.order()

        if flip == 0 and rotation % ratio == 0:
            return cn.element(rotation // ratio, 'int')
        else:
            return None

    return _map


def cn_to_dn(dn: DihedralGroup):

    def _map(e: GroupElement, dn=dn):
        assert isinstance(e.group, escnn.group.CyclicGroup)

        ratio = dn.rotation_order // e.group.order()

        return dn.element(
            (0, e.to('int') * ratio), 'int'
        )

    return _map


# Flip wrt an axis ######################################


def dn_to_flip(axis: int, flip: escnn.group.CyclicGroup):
    assert isinstance(flip, escnn.group.CyclicGroup) and flip.order() == 2

    def _map(e: GroupElement, flip=flip, axis=axis):
        assert isinstance(e.group, DihedralGroup)

        f, rot = e.to('int')

        if f == 0 and rot == 0:
            return flip.identity
        elif f == 1 and rot == axis:
            return flip.element(1)
        else:
            return None

    return _map


def flip_to_dn(axis: int, dn: DihedralGroup):

    def _map(e: GroupElement, axis=axis, dn=dn):
        assert isinstance(e.group, escnn.group.CyclicGroup) and e.group.order() == 2

        f = e.to('int')

        if f == 0:
            return dn.identity
        else:
            return dn.element((1, axis))

    return _map


# Dihedral Group ######################################


def dn_to_dm(axis: int, dm: escnn.group.DihedralGroup):
    assert isinstance(dm, escnn.group.DihedralGroup)

    def _map(e: GroupElement, dm=dm, axis=axis):
        assert isinstance(e.group, DihedralGroup)

        f, rot = e.to('int')

        ratio = e.group.rotation_order // dm.rotation_order
        if (rot - f*axis) % ratio != 0:
            return None
        else:
            return dm.element((f, (rot - f*axis) // ratio), 'int')

    return _map


def dm_to_dn(axis: int, dn: DihedralGroup):

    def _map(e: GroupElement, axis=axis, dn=dn):
        assert isinstance(e.group, escnn.group.DihedralGroup)

        f, rot = e.to('int')

        ratio = dn.rotation_order // e.group.rotation_order

        return dn.element((f, rot * ratio + f * axis), 'int')

    return _map







from escnn.gspaces import GSpace0D
from escnn.nn import init
from escnn.nn import FieldType
from escnn.nn import GeometricTensor
from .equivariant_module import EquivariantModule

from escnn.nn.modules.basismanager import BasisManager
from escnn.nn.modules.basismanager import BlocksBasisExpansion

from torch.nn import Parameter
from torch.nn.functional import linear
import torch
import numpy as np

from typing import Tuple


__all__ = ["Linear"]


class Linear(EquivariantModule):

    def __init__(self,
                 in_type: FieldType,
                 out_type: FieldType,
                 bias: bool = True,
                 basisexpansion: str = 'blocks',
                 recompute: bool = False,
                 initialize: bool = True,
                 ):
        r"""
        
        
        G-equivariant linear transformation mapping between the input and output :class:`~escnn.nn.FieldType` s
        specified by the parameters ``in_type`` and ``out_type``.
        This operation is equivariant under the action of the :attr:`escnn.nn.FieldType.fibergroup` :math:`G` of
        ``in_type`` and ``out_type``.
        
        Specifically, let :math:`\rho_\text{in}: G \to \GL{\R^{c_\text{in}}}` and
        :math:`\rho_\text{out}: G \to \GL{\R^{c_\text{out}}}` be the representations specified by the input and output
        field types.
        Then :class:`~escnn.nn.Linear` guarantees an equivariant mapping
        
        .. math::
            W \rho_\text{in}(g) v = \rho_\text{out}(g) W v \qquad\qquad \forall g \in G, u \in \R^{c_\text{in}}
            
        where :math:`\rho_\text{in}` and :math:`\rho_\text{out}` are the :math:`G`-representations associated with
        ``in_type`` and ``out_type``.

        The equivariance of a G-equivariant linear layer is guaranteed by restricting the space of weight matrices to an
        equivariant subspace.

        During training, in each forward pass the module expands the basis of G-equivariant matrices with learned weights
        before performing the linear trasformation.
        When :meth:`~torch.nn.Module.eval()` is called, the matrix is built with the current trained weights and stored
        for future reuse such that no overhead of expanding the matrix remains.
        
        .. warning ::
            
            When :meth:`~torch.nn.Module.train()` is called, the attributes :attr:`~escnn.nn.Linear.matrix` and
            :attr:`~escnn.nn.Linear.expanded_bias` are discarded to avoid situations of mismatch with the
            learnable expansion coefficients.
            See also :meth:`escnn.nn.Linear.train`.
            
            This behaviour can cause problems when storing the :meth:`~torch.nn.Module.state_dict` of a model while in
            a mode and lately loading it in a model with a different mode, as the attributes of the class change.
            To avoid this issue, we recommend converting the model to eval mode before storing or loading the state
            dictionary.
        
        .. warning ::
            To ensure compatibility with both :class:`torch.nn.Linear` and :class:`~escnn.nn.GeometricTensor`, this
            module supports only input tensors with two dimensions ``(batch_size, number_features)``.
 
 
        The learnable expansion coefficients of the this module can be initialized with the methods in
        :mod:`escnn.nn.init`.
        By default, the weights are initialized in the constructors using :func:`~escnn.nn.init.generalized_he_init`.
        
        .. warning ::
            
            This initialization procedure can be extremely slow for wide layers.
            In case initializing the model is not required (e.g. before loading the state dict of a pre-trained model)
            or another initialization method is preferred (e.g. :func:`~escnn.nn.init.deltaorthonormal_init`), the
            parameter ``initialize`` can be set to ``False`` to avoid unnecessary overhead.
        
        
        Args:
            in_type (FieldType): the type of the input field, specifying its transformation law
            out_type (FieldType): the type of the output field, specifying its transformation law
            bias (bool, optional): Whether to add a bias to the output (only to fields which contain a
                    trivial irrep) or not. Default ``True``
            basisexpansion (str, optional): the basis expansion algorithm to use. You can ignore this attribute.
            recompute (bool, optional): if ``True``, recomputes a new basis for the equivariant kernels.
                    By Default (``False``), it  caches the basis built or reuse a cached one, if it is found.
            initialize (bool, optional): initialize the weights of the model. Default: ``True``
        
        Attributes:
            
            ~.weights (torch.Tensor): the learnable parameters which are used to expand the matrix
            ~.matrix (torch.Tensor): the matrix obtained by expanding the parameters in :attr:`~escnn.nn.Linear.weights`
            ~.bias (torch.Tensor): the learnable parameters which are used to expand the bias, if ``bias=True``
            ~.expanded_bias (torch.Tensor): the equivariant bias which is summed to the output, obtained by expanding
                                    the parameters in :attr:`~escnn.nn.Linear.bias`
        
        """

        # only GSpace0D allowed since Linear acts on the last dimension of a tensor
        assert isinstance(in_type.gspace, GSpace0D)
        
        assert in_type.gspace == out_type.gspace

        super(Linear, self).__init__()
        
        self.in_type = in_type
        self.out_type = out_type
        self.space = in_type.gspace

        if bias:
            # bias can be applied only to trivial irreps inside the representation
            # to apply bias to a field we learn a bias for each trivial irreps it contains
            # and, then, we transform it with the change of basis matrix to be able to apply it to the whole field
            # this is equivalent to transform the field to its irreps through the inverse change of basis,
            # sum the bias only to the trivial irrep and then map it back with the change of basis
    
            # count the number of trivial irreps
            trivials = 0
            for r in self.out_type:
                for irr in r.irreps:
                    if self.out_type.fibergroup.irrep(*irr).is_trivial():
                        trivials += 1
    
            # if there is at least 1 trivial irrep
            if trivials > 0:
        
                # matrix containing the columns of the change of basis which map from the trivial irreps to the
                # field representations. This matrix allows us to map the bias defined only over the trivial irreps
                # to a bias for the whole field more efficiently
                bias_expansion = torch.zeros(self.out_type.size, trivials)
        
                p, c = 0, 0
                for r in self.out_type:
                    pi = 0
                    for irr in r.irreps:
                        irr = self.out_type.fibergroup.irrep(*irr)
                        if irr.is_trivial():
                            bias_expansion[p:p + r.size, c] = torch.tensor(r.change_of_basis[:, pi])
                            c += 1
                        pi += irr.size
                    p += r.size
        
                self.register_buffer("bias_expansion", bias_expansion)
                self.bias = Parameter(torch.zeros(trivials), requires_grad=True)
                self.register_buffer("expanded_bias", torch.zeros(out_type.size))
            else:
                self.bias = None
                self.expanded_bias = None
        else:
            self.bias = None
            self.expanded_bias = None

        # BasisExpansion: submodule which takes care of building the matrix
        self._basisexpansion = None
        
        if basisexpansion == 'blocks':
            self._basisexpansion = BlocksBasisExpansion(in_type.representations, out_type.representations,
                                                        self.space.build_fiber_intertwiner_basis,
                                                        np.zeros((1, 1)),
                                                        recompute=recompute)

        else:
            raise ValueError('Basis Expansion algorithm "%s" not recognized' % basisexpansion)

        self.weights = Parameter(torch.zeros(self.basisexpansion.dimension()), requires_grad=True)

        filter_size = (out_type.size, in_type.size)
        self.register_buffer("matrix", torch.zeros(*filter_size))
        if initialize:
            # by default, the weights are initialized with a generalized form of He's weight initialization
            init.generalized_he_init(self.weights.data, self.basisexpansion)
    
    def forward(self, input: GeometricTensor):
        r"""
        Convolve the input with the expanded matrix and bias.
        
        Args:
            input (GeometricTensor): input feature field transforming according to ``in_type``

        Returns:
            output feature field transforming according to ``out_type``
            
        """
        
        assert input.type == self.in_type
        # only GSpace0D allowed in practice
        assert len(input.shape) == 2
        
        if not self.training:
            _matrix = self.matrix
            _bias = self.expanded_bias
        else:
            # retrieve the matrix and the bias
            _matrix, _bias = self.expand_parameters()
        
        output = linear(input.tensor, _matrix, bias=_bias)
        
        return GeometricTensor(output, self.out_type, input.coords)

    @property
    def basisexpansion(self) -> BasisManager:
        r"""
        Submodule which takes care of building the matrix.

        It uses the learnt ``weights`` to expand a basis and returns a matrix in the usual form used by conventional
        linear modules.
        It uses the learned ``weights`` to expand the kernel in the G-steerable basis and returns it in the shape
        :math:`(c_\text{out}, c_\text{in})`.

        """
        return self._basisexpansion

    def expand_parameters(self) -> Tuple[torch.Tensor, torch.Tensor]:
        r"""

        Expand the matrix in terms of the :attr:`escnn.nn.Linear.weights` and the
        expanded bias in terms of :class:`escnn.nn.Linear.bias`.

        Returns:
            the expanded matrix and bias

        """
        _matrix = self.basisexpansion(self.weights)
        _matrix = _matrix.reshape(_matrix.shape[0], _matrix.shape[1])
    
        if self.bias is None:
            _bias = None
        else:
            _bias = self.bias_expansion @ self.bias
    
        return _matrix, _bias

    def train(self, mode=True):
        r"""

        If ``mode=True``, the method sets the module in training mode and discards the :attr:`~escnn.nn.Linear.matrix`
        and :attr:`~escnn.nn.Linear.expanded_bias` attributes.

        If ``mode=False``, it sets the module in evaluation mode. Moreover, the method builds the matrix and the bias
        using the current values of the trainable parameters and store them in :attr:`~escnn.nn.Linear.matrix` and
        :attr:`~escnn.nn.Linear.expanded_bias` such that they are not recomputed at each forward pass.

        .. warning ::

            This behaviour can cause problems when storing the :meth:`~torch.nn.Module.state_dict` of a model while in
            a mode and lately loading it in a model with a different mode, as the attributes of this class change.
            To avoid this issue, we recommend converting the model to eval mode before storing or loading the state
            dictionary.

        Args:
            mode (bool, optional): whether to set training mode (``True``) or evaluation mode (``False``).
                                   Default: ``True``.

        """
    
        if not mode:
            _matrix, _bias = self.expand_parameters()
        
            self.register_buffer("matrix", _matrix)
            if _bias is not None:
                self.register_buffer("expanded_bias", _bias)
            else:
                self.expanded_bias = None
    
        else:
            # TODO thoroughly check this is not causing problems
            if hasattr(self, "matrix"):
                del self.matrix
            if hasattr(self, "expanded_bias"):
                del self.expanded_bias
    
        return super(Linear, self).train(mode)

    def evaluate_output_shape(self, input_shape: Tuple) -> Tuple:
        assert len(input_shape) == 2
        assert input_shape[1] == self.in_type.size
        
        return (input_shape[0], self.out_type.size)

    def export(self) -> torch.nn.Linear:
        r"""
           Export this module to a normal PyTorch :class:`torch.nn.Linear` module and set to "eval" mode.

        """
        # set to eval mode so the matrix and the bias are updated with the current
        # values of the weights
        
        self.eval()
        _matrix = self.matrix
        _bias = self.expanded_bias
        
        has_bias = self.bias is not None

        # build the PyTorch module
        linear = torch.nn.Linear(self.in_type.size, self.out_type.size, bias=has_bias)
        linear.weight.data[:] = self

        # set the weights and the bias
        linear.weight.data = _matrix.data
        if has_bias:
            linear.bias.data = _bias.data

        return linear

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-6, assertion: bool = True, verbose: bool = True):
    
        x = torch.randn(10, self.in_type.size)
        x = GeometricTensor(x, self.in_type)
    
        errors = []

        for el in self.space.testing_elements:
            out1 = self(x).transform_fibers(el).tensor.detach().numpy()
            out2 = self(x.transform_fibers(el)).tensor.detach().numpy()
        
            errs = np.abs(out1 - out2)
        
            esum = np.maximum(np.abs(out1), np.abs(out2))
            esum[esum == 0.0] = 1
        
            relerr = errs / esum
        
            if verbose:
                print(el, relerr.max(), relerr.mean(), relerr.var(), errs.max(), errs.mean(), errs.var())
        
            tol = rtol * esum + atol
        
            if np.any(errs > tol) and verbose:
                print(out1[errs > tol])
                print(out2[errs > tol])
                print(tol[errs > tol])
        
            if assertion:
                assert np.all(
                    errs < tol), 'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}'.format(
                    el, errs.max(), errs.mean(), errs.var())
        
            errors.append((el, errs.mean()))
    
        return errors



from __future__ import annotations

import escnn.group
from escnn.group import Representation, GroupElement, Group
from escnn.group._numerical import decompose_representation_finitegroup
from escnn.group._numerical import decompose_representation_general

from typing import Callable, Any, List, Union, Dict, Tuple, Type

import numpy as np

__all__ = [
    "IrreducibleRepresentation",
    "build_irrep_from_generators",
    "generate_irrep_matrices_from_generators",
    "restrict_irrep"
]


class IrreducibleRepresentation(Representation):
    
    def __init__(self,
                 group: escnn.group.Group,
                 id: Tuple,
                 name: str,
                 representation: Union[Dict[escnn.group.GroupElement, np.ndarray], Callable[[Any], np.ndarray]],
                 size: int,
                 type: str,
                 supported_nonlinearities: List[str],
                 character: Union[Dict[escnn.group.GroupElement, float], Callable[[Any], float]] = None,
                 **kwargs
                 ):
        """
        Describes an "*irreducible representation*" (*irrep*).
        It is a subclass of a :class:`~escnn.group.Representation`.
        
        Irreducible representations are the building blocks into which any other representation decomposes under a
        change of basis.
        Indeed, any :class:`~escnn.group.Representation` is internally decomposed into a direct sum of irreps.
        
        Args:
            group (Group): the group which is being represented
            id (tuple): args to generate this irrep using ``group.irrep(*id)``
            name (str): an identification name for this representation
            representation (dict or callable): a callable implementing this representation or a dict mapping
                    each group element to its representation.
            size (int): the size of the vector space where this representation is defined (i.e. the size of the matrices)
            type (str): type of the irrep. It needs to be one of `R`, `C` or `H`, which represent respectively
                        real, complex and quaternionic types.
                        NOTE: this parameter substitutes the old `sum_of_squares_constituents` from *e2cnn*.
            supported_nonlinearities (list): list of nonlinearitiy types supported by this representation.
            character (callable or dict, optional): a callable returning the character of this representation for an
                    input element or a dict mapping each group element to its character.
            **kwargs: custom attributes the user can set and, then, access from the dictionary
                    in :attr:`escnn.group.Representation.attributes`
        
        Attributes:
            ~.id (tuple): tuple which identifies this irrep; it can be used to generate this irrep as ``group.irrep(*id)``
            ~.sum_of_squares_constituents (int): the sum of the squares of the multiplicities of pairwise distinct
                    irreducible constituents of the character of this representation over a non-splitting field (see
                    `Character Orthogonality Theorem <https://groupprops.subwiki.org/wiki/Character_orthogonality_theorem#Statement_over_general_fields_in_terms_of_inner_product_of_class_functions>`_
                    over general fields).
                    This attribute is fully determined by the irrep's `type` as:
                    
                    +----------+---------------------------------+
                    |  `type`  |  `sum_of_squares_constituents`  |
                    +==========+=================================+
                    |  'R'     |    `1`                          |
                    +----------+---------------------------------+
                    |  'C'     |    `2`                          |
                    +----------+---------------------------------+
                    |  'H'     |    `4`                          |
                    +----------+---------------------------------+
            
        """
        
        assert type in {'R', 'C', 'H'}

        if type == 'C':
            assert size % 2 == 0
        elif type == 'H':
            assert size % 4 == 0

        super(IrreducibleRepresentation, self).__init__(group,
                                                        name,
                                                        [id],
                                                        np.eye(size),
                                                        supported_nonlinearities,
                                                        representation=representation,
                                                        character=character,
                                                        **kwargs)
        assert isinstance(id, tuple)
        self.id: tuple = id
        self.irreducible = True
        self.type = type
        
        if self.type == 'R':
            self.sum_of_squares_constituents = 1
        elif self.type == 'C':
            self.sum_of_squares_constituents = 2
        elif self.type == 'H':
            self.sum_of_squares_constituents = 4
        else:
            raise ValueError()

    def endomorphism_basis(self) -> np.ndarray:
        if self.type == 'R':
            return np.eye(self.size).reshape(1, self.size, self.size)
        elif self.type == 'C':
            basis = np.stack([
                np.eye(2),
                np.diag([1., -1.])[::-1]
            ], axis=0)
            return np.kron(basis, np.eye(self.size // 2))
        elif self.type == 'H':
            basis = np.stack([
                np.eye(4),
                np.diag([1., -1., 1., -1.])[::-1],
                np.array([
                    [ 0.,  0., -1.,  0.],
                    [ 0.,  0.,  0., -1.],
                    [ 1.,  0.,  0.,  0.],
                    [ 0.,  1.,  0.,  0.],
                ]),
                np.array([
                    [0., -1., 0., 0.],
                    [1., 0., 0., 0.],
                    [0., 0., 0., 1.],
                    [0., 0., -1., 0.],
                ]),
            ], axis=0)
            return np.kron(basis, np.eye(self.size // 4))
        else:
            raise ValueError()


def build_irrep_from_generators(
        group: escnn.group.Group,
        generators: List[Tuple[escnn.group.GroupElement, np.ndarray]],
        id: Tuple,
        name: str,
        type: str,
        supported_nonlinearities: List[str],
        **kwargs
) -> IrreducibleRepresentation:
    
    rho = generate_irrep_matrices_from_generators(group, generators)
    d = generators[0][1].shape[0]

    return IrreducibleRepresentation(
        group,
        id,
        name,
        rho,
        d,
        type,
        supported_nonlinearities,
        **kwargs
    )


def generate_irrep_matrices_from_generators(
        group: escnn.group.Group,
        generators: List[Tuple[escnn.group.GroupElement, np.ndarray]],
) -> List[np.ndarray]:
    assert group.order() > 0
    
    d = generators[0][1].shape[0]
    
    for g, rho_g in generators:
        assert group == g.group
        assert rho_g.shape == (d, d)
    
    elements = set()
    added = set()
    
    identity = group.identity
    
    added.add(identity)
    elements.add(identity)
    
    rho = {
        g: rho_g
        for g, rho_g in generators
    }
    rho[identity] = np.eye(d)
    
    generators = [g for g, _ in generators]
    
    while len(added) > 0:
        new = set()
        for g in generators:
            for e in added:
                if g @ e not in rho:
                    rho[g @ e] = rho[g] @ rho[e]
                
                if ~g @ e not in rho:
                    rho[~g @ e] = rho[g].T @ rho[e]
                
                new |= {g @ e, ~g @ e}
        added = new - elements
        elements |= added
    
    assert len(elements) == group.order(), 'Error! The set of generators passed does not generate the whole group'
    
    for a in elements:
        assert ~a in elements
        
        assert np.allclose(
            rho[~a],
            rho[a].T
        )
        
        for b in elements:
            assert a @ b in elements
            
            assert np.allclose(
                rho[a] @ rho[b],
                rho[a @ b]
            )
    
    return rho


from joblib import Memory
# import os
# cache = Memory(os.path.join(os.path.dirname(__file__), '_jl_restricted_irreps'), verbose=2)

from escnn.group import __cache_path__
cache = Memory(__cache_path__, verbose=0)


@cache.cache
def _restrict_irrep(irrep_id: Tuple, id, group_class: str, **group_keys) -> Tuple[np.matrix, List[Tuple[Tuple, int]]]:
    
    group = escnn.group.groups_dict[group_class]._generator(**group_keys)
    
    irrep = group.irrep(*irrep_id)

    id = irrep.group._decode_subgroup_id_pickleable(id)

    subgroup, parent, child = group.subgroup(id)
    
    if subgroup.order() == 1:
        # if the subgroup is the trivial group, just return the identity cob and the list of trivial reprs
        return np.eye(irrep.size), [subgroup.trivial_representation.id]*irrep.size
    
    if subgroup.order() > 1:
        # if it is a finite group, we can sample all the element and use the precise method based on Character theory

        representation = {
            g: irrep(parent(g)) for g in subgroup.elements
        }

        # to solve the Sylvester equation and find the change of basis matrix, it is sufficient to sample
        # the generators of the subgroup
        change_of_basis, multiplicities = decompose_representation_finitegroup(
            representation,
            subgroup,
        )

    else:
        # if the group is not finite, we rely on the numerical method which is based on some samples of the group

        representation = lambda g: irrep(parent(g))

        change_of_basis, multiplicities = decompose_representation_general(
            representation,
            subgroup,
        )

    irreps = []
    
    for irr, m in multiplicities:
        irreps += [irr]*m
    
    return change_of_basis, irreps


def restrict_irrep(irrep: IrreducibleRepresentation, id) -> Tuple[np.matrix, List[Tuple[str, int]]]:
    r"""
        Restrict the input `irrep` to the subgroup identified by `id`.
    """
    group_keys = irrep.group._keys

    id = irrep.group._encode_subgroup_id_pickleable(id)

    return _restrict_irrep(irrep.id, id, irrep.group.__class__.__name__, **group_keys)





import escnn.nn
from escnn.nn import FieldType, GeometricTensor
from escnn.group import Representation
from escnn.kernels import KernelBasis

from .rd_point_convolution import _RdPointConv

from typing import Callable, Tuple, Dict, Union, List

import torch
import numpy as np


import math


__all__ = ["R2PointConv"]


class R2PointConv(_RdPointConv):
    
    def __init__(self,
                 in_type: FieldType,
                 out_type: FieldType,
                 groups: int = 1,
                 bias: bool = True,
                 sigma: Union[List[float], float] = None,
                 width: float = None,
                 n_rings: int = None,
                 frequencies_cutoff: Union[float, Callable[[float], int]] = None,
                 rings: List[float] = None,
                 basis_filter: Callable[[dict], bool] = None,
                 recompute: bool = False,
                 initialize: bool = True,
                 ):
        r"""

        G-steerable planar convolution mapping between the input and output :class:`~escnn.nn.FieldType` s specified by
        the parameters ``in_type`` and ``out_type``.
        This operation is equivariant under the action of :math:`\R^2\rtimes G` where :math:`G` is the
        :attr:`escnn.nn.FieldType.fibergroup` of ``in_type`` and ``out_type``.

        This class implements convolution with steerable filters over sparse planar geometric graphs.
        Instead, :class:`~escnn.nn.R2Conv` implements an equivalent convolution layer over a pixel grid.
        See the documentation of :class:`~escnn.nn.R2Conv` for more details about equivariance and steerable convolution.

        The input of this module is a planar geometric graph, i.e. a graph whose nodes are associated with 2D
        coordinates in :math:`\R^2`.
        The nodes' coordinates should be stored in the ``coords`` attribute of the input
        :class:`~escnn.nn.GeometricTensor`.
        The adjacency of the graph should be passed as a second input tensor ``edge_index``, like commonly done in
        :class:`~torch_geometric.nn.conv.message_passing.MessagePassing`.
        See :meth:`~escnn.nn.modules.pointconv._RdPointConv.forward`.

        In each forward pass, the module computes the relative coordinates of the points on the edges and samples each
        filter in the basis of G-steerable kernels at these relative locations.
        The basis filters are expanded using the learnable weights and used to perform convolution over the graph in the
        message passing framework.
        Optionally, the relative coordinates can be pre-computed and passed in the input ``edge_delta`` tensor.

        .. note ::
            In practice, we first apply the basis filters on the input features and then combine the responses via the
            learnable weights. See also :meth:`~escnn.nn.modules.basismanager.BlocksBasisSampler.compute_messages`.

        .. warning ::

            When :meth:`~torch.nn.Module.eval()` is called, the bias is built with the current trained weights and stored
            for future reuse such that no overhead of expanding the bias remains.

            When :meth:`~torch.nn.Module.train()` is called, the attribute :attr:`~escnn.nn.R2PointConv.expanded_bias`
            is discarded to avoid situations of mismatch with the learnable expansion coefficients.
            See also :meth:`escnn.nn.R2PointConv.train`.

            This behaviour can cause problems when storing the :meth:`~torch.nn.Module.state_dict` of a model while in
            a mode and lately loading it in a model with a different mode, as the attributes of the class change.
            To avoid this issue, we recommend converting the model to eval mode before storing or loading the state
            dictionary.

        The learnable expansion coefficients of this module can be initialized with the methods in
        :mod:`escnn.nn.init`.
        By default, the weights are initialized in the constructors using :func:`~escnn.nn.init.generalized_he_init`.

        .. warning ::

            This initialization procedure can be *extremely* slow for wide layers.
            In case initializing the model is not required (e.g. before loading the state dict of a pre-trained model)
            or another initialization method is preferred (e.g. :func:`~escnn.nn.init.deltaorthonormal_init`), the
            parameter ``initialize`` can be set to ``False`` to avoid unnecessary overhead.

        The parameters ``width``, ``n_rings``, ``sigma``, ``frequencies_cutoff``, ``rings`` are
        optional parameters used to control how the basis for the filters is built, how it is sampled and how it is
        expanded to build the filter.
        In practice, steerable filters are parameterized independently on a number of concentric rings by using
        *circular harmonics*.
        These rings can be specified by *i)* either using the list ``rings``, which defines the radii of each ring, or
        by *ii)* indicating the maximum radius ``width`` and the number ``n_rings`` of rings to include.
        ``sigma`` defines the "thickness" of each ring as the standard deviation of a Gaussian bell along the radial
        direction.
        ``frequencies_cutoff`` is a function defining the maximum frequency of the circular harmonics allowed at each
        radius. If a `float` value is passed, the maximum frequency is equal to the radius times this factor.

        .. note ::
            These parameters should be carefully tuned depending on the typical connectivity and the scale of the
            geometric graphs of interest.

        .. warning ::
            We don't support ``groups > 1`` yet.
            We include this parameter for future compatibility.

        Args:
            in_type (FieldType): the type of the input field, specifying its transformation law
            out_type (FieldType): the type of the output field, specifying its transformation law
            groups (int, optional): number of blocked connections from input channels to output channels.
                                    It allows depthwise convolution. When used, the input and output types need to be
                                    divisible in ``groups`` groups, all equal to each other.
                                    Default: ``1``.
            bias (bool, optional): Whether to add a bias to the output (only to fields which contain a
                    trivial irrep) or not. Default ``True``
            sigma (list or float, optional): width of each ring where the bases are sampled. If only one scalar
                    is passed, it is used for all rings.
            width (float, optional): radius of the support of the learnable filters.
                    Setting ``n_rings`` and ``width`` is an alternative to use ``rings``.
            n_rings (int, optional): number of (equally spaced) rings the support of the filters is split into.
                    Setting ``n_rings`` and ``width`` is an alternative to use ``rings``.
            frequencies_cutoff (callable or float, optional): function mapping the radii of the basis elements to the
                    maximum frequency accepted. If a float values is passed, the maximum frequency is equal to the
                    radius times this factor. Default: ``3.``.
            rings (list, optional): radii of the rings where to sample the bases
            basis_filter (callable, optional): function which takes as input a descriptor of a basis element
                    (as a dictionary) and returns a boolean value: whether to preserve (``True``) or discard (``False``)
                    the basis element. By default (``None``), no filtering is applied.
            recompute (bool, optional): if ``True``, recomputes a new basis for the equivariant kernels.
                    By Default (``False``), it  caches the basis built or reuse a cached one, if it is found.
            initialize (bool, optional): initialize the weights of the model. Default: ``True``

        Attributes:

            ~.weights (torch.Tensor): the learnable parameters which are used to expand the kernel
            ~.bias (torch.Tensor): the learnable parameters which are used to expand the bias, if ``bias=True``
            ~.expanded_bias (torch.Tensor): the equivariant bias which is summed to the output, obtained by expanding
                                    the parameters in :attr:`~escnn.nn.R2PointConv.bias`

        """

        basis_filter, self._rings, self._sigma, self._maximum_frequency = compute_basis_params(
            frequencies_cutoff, rings, sigma, width, n_rings, basis_filter
        )

        super(R2PointConv, self).__init__(
            in_type, out_type,
            d=2,
            groups=groups,
            bias=bias,
            basis_filter=basis_filter,
            recompute=recompute
        )

        if initialize:
            # by default, the weights are initialized with a generalized form of He's weight initialization
            escnn.nn.init.generalized_he_init(self.weights.data, self.basissampler)

    def _build_kernel_basis(self, in_repr: Representation, out_repr: Representation) -> KernelBasis:
        return self.space.build_kernel_basis(in_repr, out_repr,
                                             self._sigma, self._rings,
                                             maximum_frequency=self._maximum_frequency
                                             )


def bandlimiting_filter(frequency_cutoff: Union[float, Callable[[float], float]]) -> Callable[[dict], bool]:
    r"""

    Returns a method which takes as input the attributes (as a dictionary) of a basis element and returns a boolean
    value: whether to preserve that element (true) or not (false)

    If the parameter ``frequency_cutoff`` is a scalar value, the maximum frequency allowed at a certain radius is
    proportional to the radius itself. in thi case, the parameter ``frequency_cutoff`` is the factor controlling this
    proportionality relation.

    If the parameter ``frequency_cutoff`` is a callable, it needs to take as input a radius (a scalar value) and return
    the maximum frequency which can be sampled at that radius.

    args:
        frequency_cutoff (float): factor controlling the bandlimiting

    returns:
        a function which checks the attributes of individual basis elements and chooses whether to discard them or not

    """

    if isinstance(frequency_cutoff, float):
        frequency_cutoff = lambda r, fco=frequency_cutoff: r * frequency_cutoff

    def bl_filter(attributes: dict) -> bool:
        return math.fabs(attributes["irrep:frequency"]) <= frequency_cutoff(attributes["radius"])

    return bl_filter


def compute_basis_params(
        frequencies_cutoff: Union[float, Callable[[float], float]] = None,
        rings: List[float] = None,
        sigma: List[float] = None,
        width: float = None,
        n_rings: int = None,
        custom_basis_filter: Callable[[dict], bool] = None,
):

    assert (width is not None and n_rings is not None) != (rings is not None)

    # by default, the number of rings equals half of the filter size
    if rings is None:
        assert width > 0.
        assert n_rings > 0
        rings = torch.linspace(0, width, n_rings)
        rings = rings.tolist()

    if sigma is None:
        sigma = [0.6] * (len(rings) - 1) + [0.4]
        for i, r in enumerate(rings):
            if r == 0.:
                sigma[i] = 0.005
    elif isinstance(sigma, float):
        sigma = [sigma] * len(rings)

    if frequencies_cutoff is None:
        frequencies_cutoff = 3.

    if isinstance(frequencies_cutoff, float):
        frequencies_cutoff = lambda r, fco=frequencies_cutoff: fco * r

    # check if the object is a callable function
    assert callable(frequencies_cutoff)

    maximum_frequency = int(max(frequencies_cutoff(r) for r in rings))

    fco_filter = bandlimiting_filter(frequencies_cutoff)

    if custom_basis_filter is not None:
        basis_filter = lambda d, custom_basis_filter=custom_basis_filter, fco_filter=fco_filter: (
                custom_basis_filter(d) and fco_filter(d)
        )
    else:
        basis_filter = fco_filter

    return basis_filter, rings, sigma, maximum_frequency





from typing import List, Tuple, Any

from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule

import torch
from torch.nn import BatchNorm3d, BatchNorm2d

from ..utils import indexes_from_labels

__all__ = ["InnerBatchNorm"]


class BatchNorm4d(torch.nn.modules.batchnorm._BatchNorm):
    def _check_input_dim(self, input):
        if input.dim() != 6:
            raise ValueError(
                "expected 6D input (got {}D input)".format(input.dim())
            )


class InnerBatchNorm(EquivariantModule):
    
    def __init__(self,
                 in_type: FieldType,
                 eps: float = 1e-05,
                 momentum: float = 0.1,
                 affine: bool = True,
                 track_running_stats: bool = True,
                 ):
        r"""
        
        Batch normalization for representations with permutation matrices.
        
        Statistics are computed both over the batch and the spatial dimensions and over the channels within
        the same field (which are permuted by the representation).
        
        Only representations supporting pointwise non-linearities are accepted as input field type.
        
        Args:
            in_type (FieldType): the input field type
            eps (float, optional): a value added to the denominator for numerical stability. Default: ``1e-5``
            momentum (float, optional): the value used for the ``running_mean`` and ``running_var`` computation.
                    Can be set to ``None`` for cumulative moving average (i.e. simple average). Default: ``0.1``
            affine (bool, optional):  if ``True``, this module has learnable affine parameters. Default: ``True``
            track_running_stats (bool, optional): when set to ``True``, the module tracks the running mean and variance;
                                                  when set to ``False``, it does not track such statistics but uses
                                                  batch statistics in both training and eval modes.
                                                  Default: ``True``
            
        """

        assert isinstance(in_type.gspace, GSpace)
        
        super(InnerBatchNorm, self).__init__()

        for r in in_type.representations:
            assert 'pointwise' in r.supported_nonlinearities, \
                'Error! Representation "{}" does not support "pointwise" non-linearity'.format(r.name)

        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type
        
        self.affine = affine
        self.eps = eps
        self.momentum = momentum
        self.track_running_stats = track_running_stats
        
        # group fields by their size and
        #   - check if fields with the same size are contiguous
        #   - retrieve the indices of the fields
        grouped_fields = indexes_from_labels(self.in_type, [r.size for r in self.in_type.representations])

        # number of fields of each size
        self._nfields = {}
        
        # indices of the channels corresponding to fields belonging to each group
        _indices = {}
        
        # whether each group of fields is contiguous or not
        self._contiguous = {}
        
        for s, (contiguous, fields, indices) in grouped_fields.items():
            self._nfields[s] = len(fields)
            self._contiguous[s] = contiguous
            
            if contiguous:
                # for contiguous fields, only the first and last indices are kept
                _indices[s] = torch.LongTensor([min(indices), max(indices)+1])
            else:
                # otherwise, transform the list of indices into a tensor
                _indices[s] = torch.LongTensor(indices)
                
            # register the indices tensors as parameters of this module
            self.register_buffer('indices_{}'.format(s), _indices[s])

        if self.in_type.gspace.dimensionality in [0, 1]:
            bnorm_class = torch.nn.BatchNorm2d
        elif self.in_type.gspace.dimensionality == 2:
            bnorm_class = torch.nn.BatchNorm3d
        elif self.in_type.gspace.dimensionality == 3:
            bnorm_class = BatchNorm4d
        else:
            raise NotImplementedError

        for s in _indices.keys():
            _batchnorm = bnorm_class(
                self._nfields[s],
                self.eps,
                self.momentum,
                affine=self.affine,
                track_running_stats=self.track_running_stats
            )
            self.add_module('batch_norm_[{}]'.format(s), _batchnorm)

    def reset_running_stats(self):
        for s, contiguous in self._contiguous.items():
            batchnorm = getattr(self, f'batch_norm_[{s}]')
            batchnorm.reset_running_stats()

    def reset_parameters(self):
        for s, contiguous in self._contiguous.items():
            batchnorm = getattr(self, f'batch_norm_[{s}]')
            batchnorm.reset_parameters()

    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        
        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map
            
        """
        
        assert input.type == self.in_type
        
        b, c = input.tensor.shape[:2]
        shape = input.tensor.shape[2:]

        output = torch.empty_like(input.tensor)
        
        # iterate through all field sizes
        for s, contiguous in self._contiguous.items():
            
            indices = getattr(self, f"indices_{s}")
            batchnorm = getattr(self, f'batch_norm_[{s}]')
            
            if contiguous:
                # if the fields were contiguous, we can use slicing
                output[:, indices[0]:indices[1], :, :] = batchnorm(
                    input.tensor[:, indices[0]:indices[1], :, :].view(b, -1, s, *shape)
                ).view(b, -1, *shape)
            else:
                # otherwise we have to use indexing
                output[:, indices, :, :] = batchnorm(
                    input.tensor[:, indices, :, :].view(b, -1, s, *shape)
                ).view(b, -1, *shape)
        
        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, input.coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:

        assert len(input_shape) >= 2
        assert input_shape[1] == self.in_type.size

        b, c = input_shape[:2]
        spatial_shape = input_shape[2:]

        return (b, self.out_type.size, *spatial_shape)

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
        # return super(InnerBatchNorm, self).check_equivariance(atol=atol, rtol=rtol)
        pass

    def export(self):
        r"""
        Export this module to a normal PyTorch :class:`torch.nn.BatchNorm2d` module and set to "eval" mode.

        """

        if not self.track_running_stats:
            raise ValueError('''
                Equivariant Batch Normalization can not be converted into conventional batch normalization when
                "track_running_stats" is False because the statistics contained in a single batch are generally
                not symmetric
            ''')
        
        self.eval()

        if self.in_type.gspace.dimensionality in [0, 1]:
            bnorm_class = torch.nn.BatchNorm1d
        elif self.in_type.gspace.dimensionality == 2:
            bnorm_class = torch.nn.BatchNorm2d
        elif self.in_type.gspace.dimensionality == 3:
            bnorm_class = torch.nn.BatchNorm3d
        else:
            raise NotImplementedError

        batchnorm = bnorm_class(
            self.in_type.size,
            self.eps,
            self.momentum,
            affine=self.affine,
            track_running_stats=self.track_running_stats
        )
        
        num_batches_tracked = None
        
        for s, contiguous in self._contiguous.items():
            if not contiguous:
                raise NotImplementedError(
                    '''Non-contiguous indices not supported yet when converting
                    inner-batch normalization into conventional BatchNorm2d'''
                )
            
            # indices = getattr(self, 'indices_{}'.format(s))
            start, end = getattr(self, 'indices_{}'.format(s))
            bn = getattr(self, 'batch_norm_[{}]'.format(s))
            
            n = self._nfields[s]
            
            batchnorm.running_var.data[start:end] = bn.running_var.data.view(n, 1).expand(n, s).reshape(-1)
            batchnorm.running_mean.data[start:end] = bn.running_mean.data.view(n, 1).expand(n, s).reshape(-1)
            batchnorm.num_batches_tracked.data = bn.num_batches_tracked.data

            if num_batches_tracked is None:
                num_batches_tracked = bn.num_batches_tracked.data
            else:
                assert num_batches_tracked == bn.num_batches_tracked.data
            
            if self.affine:
                batchnorm.weight.data[start:end] = bn.weight.data.view(n, 1).expand(n, s).reshape(-1)
                batchnorm.bias.data[start:end] = bn.bias.data.view(n, 1).expand(n, s).reshape(-1)

        batchnorm.eval()

        return batchnorm

    def __repr__(self):
        extra_lines = []
        extra_repr = self.extra_repr()
        if extra_repr:
            extra_lines = extra_repr.split('\n')
    
        main_str = self._get_name() + '('
        if len(extra_lines) == 1:
            main_str += extra_lines[0]
        else:
            main_str += '\n  ' + '\n  '.join(extra_lines) + '\n'
    
        main_str += ')'
        return main_str

    def extra_repr(self):
        return '{in_type}, eps={eps}, momentum={momentum}, affine={affine}, track_running_stats={track_running_stats}'\
            .format(**self.__dict__)




from __future__ import annotations

from escnn.group import change_basis, directsum
from escnn.group.irrep import generate_irrep_matrices_from_generators
from escnn.group.irrep import restrict_irrep
from escnn.group.utils import cycle_isclose

from .utils import *

from .so3_utils import PARAMETRIZATION as PARAMETRIZATION_SO3
from .so3_utils import PARAMETRIZATIONS
from .so3_utils import IDENTITY, _grid, _combine, _equal, _invert, _change_param, _check_param, _hash

from .so3group import _build_character, _build_irrep

import numpy as np

from typing import Tuple, Callable, Iterable, List, Dict, Any, Union


__all__ = ["Octahedral"]

_PHI = (1. + np.sqrt(5)) / 2


class Octahedral(Group):

    PARAM = PARAMETRIZATION_SO3
    
    PARAMETRIZATIONS = PARAMETRIZATIONS
    
    def __init__(self):
        r"""

        Subgroup Structure:

        +-----------------------------------+-----------------------------------+-------------------------------------------------------------------------------------------------------------------+
        |    ``id[0]``                      |    ``id[1]``                      |    subgroup                                                                                                       |
        +===================================+===================================+===================================================================================================================+
        |        'octa'                     |                                   |   The Octahedral :math:`O` group itself                                                                           |
        +-----------------------------------+-----------------------------------+-------------------------------------------------------------------------------------------------------------------+
        |        'tetra'                    |                                   |   Tetrahedral :math:`T` subgroup                                                                                  |
        +-----------------------------------+-----------------------------------+-------------------------------------------------------------------------------------------------------------------+
        |        False                      |     N = 1, 2, 3, 4                |   :math:`C_N` of N discrete planar rotations                                                                      |
        +-----------------------------------+-----------------------------------+-------------------------------------------------------------------------------------------------------------------+
        |        True                       |     N = 2, 3, 4                   |   *dihedral* :math:`D_N` subgroup of N discrete planar rotations and out-of-plane :math:`\pi` rotation            |
        +-----------------------------------+-----------------------------------+-------------------------------------------------------------------------------------------------------------------+
        |        True                       |     1                             |   equivalent to ``(False, 2, adj)``                                                                               |
        +-----------------------------------+-----------------------------------+-------------------------------------------------------------------------------------------------------------------+


        """
        
        super(Octahedral, self).__init__("Octahedral", False, False)
        
        self._identity = self.element(IDENTITY)
        
        self._elements = [self.element(g) for g in _grid('cube')]
        assert len(self._elements) == 24
            
        # self._identity = self._elements[3]

        self._generators = [
            self._elements[17],  # Cyclic Group of order 4
            self._elements[11],  # Cyclic Group of order 3
            self._elements[22],  # Cyclic Group of order 2
        ]

        self._build_representations()

    @property
    def generators(self) -> List[GroupElement]:
        return self._generators

    @property
    def identity(self) -> GroupElement:
        return self._identity

    @property
    def elements(self) -> List[GroupElement]:
        return self._elements
     
    @property
    def _keys(self) -> Dict[str, Any]:
        return dict()

    @property
    def subgroup_trivial_id(self):
        raise NotImplementedError

    @property
    def subgroup_self_id(self):
        raise NotImplementedError
        return 'octa'

    ###########################################################################
    # METHODS DEFINING THE GROUP LAW AND THE OPERATIONS ON THE GROUP'S ELEMENTS
    ###########################################################################

    def _inverse(self, element, param=PARAM):
        r"""
        Return the inverse element of the input element
        """
        return _invert(element, param=param)

    def _combine(self, e1, e2,
                param=PARAM,
                param1=None,
                param2=None
                ):
        r"""
        Return the sum of the two input elements
        """
        return _combine(e1, e2, param=param, param1=param1, param2=param2)

    def _equal(self, e1, e2,
              param=PARAM,
              param1=None,
              param2=None,
              ) -> bool:
        r"""
        Check if the two input values corresponds to the same element
        """
        return _equal(e1, e2, param=param, param1=param1, param2=param2)

    def _hash_element(self, element, param=PARAM):
        return _hash(element, param)

    def _repr_element(self, element, param=PARAM):
        return element.__repr__()

    def _is_element(self, element,
                    param: str = PARAM,
                    verbose: bool = False,
                    ) -> bool:

        ATOL = 1e-7
        RTOL = 1e-5

        if not _check_param(element, param):
            if verbose:
                print(f"Element {element} is not a rotation")
            return False

        # convert to matrix representation
        element = self._change_param(element, param, 'MAT')

        # take absolute value of the elements
        # note that we have already ensured that the determinant is positive using `_check_param` above since it checks
        # that it is a rotation
        at = np.abs(element)

        # check if the matrix is a permutation matrix
        ans = (
                np.isclose(at.sum(axis=0), 1., atol=ATOL, rtol=RTOL).all()
            and np.isclose(at.sum(axis=1), 1., atol=ATOL, rtol=RTOL).all()
            and (np.isclose(at, 1., atol=ATOL, rtol=RTOL) | np.isclose(at, 0., atol=ATOL, rtol=RTOL)).all()
        )

        return ans

    def _change_param(self, element, p_from: str, p_to: str):
        assert p_from in self.PARAMETRIZATIONS
        assert p_to in self.PARAMETRIZATIONS
        return _change_param(element, p_from, p_to)

    ###########################################################################

    def sample(self, param: str = PARAM) -> GroupElement:
        return self._elements[
            np.random.randint(self.order())
        ]

    def testing_elements(self) -> Iterable[GroupElement]:
    
        r"""
        A finite number of group elements to use for testing.
        """
        return iter(self._elements)

    def __eq__(self, other):
        if not isinstance(other, Octahedral):
            return False
        else:
            return self.name == other.name

    def _process_subgroup_id(self, id):

        if not isinstance(id, tuple):
            id = (id,)

        assert isinstance(id[0], bool) or isinstance(id[0], str), id[0]

        if not isinstance(id[-1], GroupElement):
            id = (*id, self.identity)

        assert id[-1].group == self

        if isinstance(id[0], bool):
            assert id[1] in [1, 2, 3, 4]

            if id[0] == True and id[1] == 1:
                # flip subgroup of the O(2) subgroup of SO(3)
                # this is equivalent to the C_2 subgroup of 180 deg rotations out of the plane (around X axis)

                V = np.array([1., 1., -1.])
                V /= np.linalg.norm(V)

                change_axis = np.zeros(4)
                change_axis[:3] = V * np.sin(np.pi/3.)
                change_axis[3] = np.cos(np.pi/3.)

                adj = self.element(change_axis, 'Q') @ id[-1]
                id = (False, 2, adj)

        return id

    def _subgroup(self, id) -> Tuple[
        Group,
        Callable[[GroupElement], GroupElement],
        Callable[[GroupElement], GroupElement]
    ]:
        r"""

        Returns:
            a tuple containing

                - the subgroup,

                - a function which maps an element of the subgroup to its inclusion in the original group and

                - a function which maps an element of the original group to the corresponding element in the subgroup (returns None if the element is not contained in the subgroup)

        """
        # TODO : implement this!

        sg = None
        parent_map = None
        child_map = None

        id, adj = id[:-1], id[-1]

        if id == ('octa',):
            sg = self
            parent_map = build_adjoint_map(self, ~adj)
            child_map = build_adjoint_map(self, adj)
        elif id == ('tetra',):
            raise NotImplementedError()
        elif id == (False, 1):
            sg = escnn.group.cyclic_group(1)
            parent_map, child_map = build_trivial_subgroup_maps(self)
        elif id == (False, 2):
            sg = escnn.group.cyclic_group(2)
            axis = np.asarray([0., 0., 1.])
            parent_map = cn_to_octa(adj, sg, axis=axis)
            child_map = octa_to_cn(adj, sg, axis=axis)
        elif id == (False, 3):
            sg = escnn.group.cyclic_group(3)
            axis = np.asarray([1., 1., 1.]) / np.sqrt(3)
            parent_map = cn_to_octa(adj, sg, axis=axis)
            child_map = octa_to_cn(adj, sg, axis=axis)
        elif id == (False, 4):
            sg = escnn.group.cyclic_group(4)
            axis = np.asarray([0., 0., 1.])
            axis /= np.linalg.norm(axis)
            parent_map = cn_to_octa(adj, sg, axis=axis)
            child_map = octa_to_cn(adj, sg, axis=axis)
        elif id == (True, 2):
            sg = escnn.group.dihedral_group(2)
            parent_map, child_map = None, None
            raise NotImplementedError()
        elif id == (True, 3):
            sg = escnn.group.dihedral_group(3)
            parent_map, child_map = None, None
            raise NotImplementedError()
        elif id == (True, 4):
            sg = escnn.group.dihedral_group(4)
            parent_map, child_map = None, None
            raise NotImplementedError()
        else:
            raise ValueError(f'Subgroup id {id} not recognized!')

        return sg, parent_map, child_map

    def _restrict_irrep(self, irrep: str, id) -> Tuple[np.matrix, List[str]]:
        r"""

        Returns:
            a pair containing the change of basis and the list of irreps of the subgroup which appear in the restricted irrep

        """
        sg_id, adj = id[:-1], id[-1]

        irr = self.irrep(*irrep)

        sg, _, _ = self.subgroup(id)

        irreps = []
        change_of_basis = None

        try:
            if sg_id == ('octa', ):
                change_of_basis = irr.change_of_basis
                irreps = irr.irreps
            elif sg_id == (False, 1):
                change_of_basis = np.eye(irr.size)
                irreps = [(1,)]*irr.size
            else:
                raise NotImplementedError()

        except NotImplementedError:
            if sg.order() > 0:
                change_of_basis, irreps = restrict_irrep(irr, sg_id)
            else:
                raise

        change_of_basis = self.irrep(*irrep)(adj).T @ change_of_basis

        return change_of_basis, irreps

    def _build_representations(self):
        r"""
        Build the irreps for this group

        """

        # Build all the Irreducible Representations

        # add Trivial representation
        self.irrep(0)

        # add other irreducible representations

        # Frequency 1 Wigner D matrix
        self.irrep(1)

        # Frequency 2 Wigner D matrix decomposes as a direct sum of a 2 and a 3 dimensional irrep
        self.irrep(-1) # 3 dimensional irrep
        self.irrep(2) # 2 dimensional irrep

        # SO(3)'s freq 3 irrep decomposes in a 1-dimensional irrep and the sum of the two previous 3 dimensional irreps
        self.irrep(3) # 1 dimensional

        # add all the irreps to the set of representations already built for this group
        self.representations.update(**{irr.name: irr for irr in self.irreps()})

        # build the regular representation

        # N.B.: it represents the LEFT-ACTION of the elements
        self.representations['regular'] = self.regular_representation

    @property
    def trivial_representation(self) -> Representation:
        return self.irrep(0)

    @property
    def standard_representation(self) -> Representation:
        r"""
        Restriction of the standard representation of SO(3) as 3x3 rotation matrices

        """
        name = f'standard'
    
        if name not in self._representations:
            change_of_basis = np.array([
                [0, 0, 1],
                [1, 0, 0],
                [0, 1, 0]
            ])
        
            self._representations[name] = change_basis(
                self.irrep(1),
                change_of_basis=change_of_basis,
                name=name,
                supported_nonlinearities=self.irrep(1).supported_nonlinearities,
            )
    
        return self._representations[name]

    @property
    def cube_vertices_representation(self) -> Representation:
        # action on the 8 vertices of the cube (or faces of the octahedron)

        sgid = (False, 3)
        return self.quotient_representation(sgid, name='cube_vertices')

    @property
    def cube_faces_representation(self) -> Representation:
        # action on the 6 faces of the cube (or vertices of the octahedron)

        sgid = (False, 4)
        return self.quotient_representation(sgid, name='cube_faces')

    @property
    def cube_edges_representation(self) -> Representation:
        # action on the 12 edges of the cube or octahedron

        sgid = (True, 1)
        return self.quotient_representation(sgid, name='cube_edges')

    def irrep(self, l: int) -> IrreducibleRepresentation:
        r"""
        Build the irrep of :math:`O` identified by the integer :math:`l`.
        For :math:`l = 0, 1`, the irrep is equivalent to the Wigner D matrix of the same frequency :math:`l`.
        For :math:`l=2`, the 5-dimensional Wigner D matrix is decomposed in a 3-dimensional and a 2-dimensional irreps,
        here identified respectively by :math:`l=-1` and :math:`l=2`.
        For :math:`l=3`, the 7-dimensional Wigner D matrix is decomposed in a 1-dimensional irrep and the two previous
        3-dimensional irreps, here identified respectively by :math:`l=3` and :math:`l=1, -1`.
        
        Args:
            l (int): identifier of the irrep

        Returns:
            the corresponding irrep

        """
        
        assert isinstance(l, int)
        assert -1 <= l <= 3

        name = f"irrep_{l}"
        id = (l,)

        if id not in self._irreps:

            if l == 0:
                # Trivial representation
                irrep = build_trivial_irrep()
                character = build_trivial_character()
                supported_nonlinearities = ['pointwise', 'norm', 'gated', 'gate']
                self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 1, 'R',
                                                              supported_nonlinearities=supported_nonlinearities,
                                                              character=character,
                                                              )
            elif l == 1:
        
                # Irreducible Representation equivalent to the frequency 1 Wigner D matrices
                irrep = _build_irrep(l)
                character = _build_character(l)
                supported_nonlinearities = ['norm', 'gated']
                self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 3, 'R',
                                                              supported_nonlinearities=supported_nonlinearities,
                                                              character=character)
            elif l == -1 or l == 2:
    
                irrep = _build_octa_irrep(self, l)
                supported_nonlinearities = ['norm', 'gated']
                self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, irrep[self.identity].shape[0], 'R',
                                                             supported_nonlinearities=supported_nonlinearities)

            elif l == 3:
                irrep = _build_octa_irrep(self, l)
                supported_nonlinearities = ['norm', 'gated', 'concatenated']
                self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, irrep[self.identity].shape[0], 'R',
                                                             supported_nonlinearities=supported_nonlinearities)

            else:
                raise ValueError()

        return self._irreps[id]

    _cached_group_instance = None

    @classmethod
    def _generator(cls) -> 'Octahedral':
        if cls._cached_group_instance is None:
            cls._cached_group_instance = Octahedral()
    
        return cls._cached_group_instance


def _is_axis_aligned(v: np.ndarray, n: int, verbose: bool = False, ATOL=1e-7, RTOL = 1e-5) -> bool:

    norm = np.linalg.norm(v)
    v = v / norm

    if n == 2:
        # rotation of order 2

        # the rotation axis need to be aligned with one of the axes X, Y, Z or to
        # the bisector of a pair of axes XY, XZ, YZ
        # There are in total 6 + 12 possible vectors

        # remove sign ambiguity
        v = np.abs(v)

        axes = np.eye(3)
        bisectors = np.array([
            [1., 0., 1.],
            [1., 1., 0.],
            [0., 1., 1.],
        ]) / np.sqrt(2)

        ans = (
                # axes aligned
                np.allclose(v, axes[0], atol=ATOL, rtol=RTOL)
                or np.allclose(v, axes[1], atol=ATOL, rtol=RTOL)
                or np.allclose(v, axes[2], atol=ATOL, rtol=RTOL)
                # bisectors aligned
                or np.allclose(v, bisectors[0], atol=ATOL, rtol=RTOL)
                or np.allclose(v, bisectors[1], atol=ATOL, rtol=RTOL)
                or np.allclose(v, bisectors[2], atol=ATOL, rtol=RTOL)
        )

        if not ans and verbose:
            print(f'Rotation by a multiple of 2pi/{n} not aligned with a {n}-fold rotational axis of the Octahedron.')

        return ans

    elif n == 4:
        # rotation of order 4

        # the rotation axis need to be aligned with one of the axes X, Y, Z
        # There are in total 6 possible vectors

        # remove sign ambiguity
        v = np.abs(v)

        axes = np.eye(3)

        ans = (
               np.allclose(v, axes[0], atol=ATOL, rtol=RTOL)
            or np.allclose(v, axes[1], atol=ATOL, rtol=RTOL)
            or np.allclose(v, axes[2], atol=ATOL, rtol=RTOL)
        )

        if not ans and verbose:
            print(f'Rotation by a multiple of 2pi/{n} not aligned with a {n}-fold rotational axis of the Octahedron.')

        return ans

    elif n == 3:
        # rotation or order 3

        # the rotation axis need to pass through one of the vertices of the cube
        # There are in total 8 possible vectors

        # remove sign ambiguity
        v = np.abs(v)

        # since the vector is normalized, `v` should now be `(1, 1, 1)^T * 1/sqrt(3)`

        ans = np.allclose(v, 1./np.sqrt(3), atol=ATOL, rtol=RTOL)

        if not ans and verbose:
            print(f'Rotation by a multiple of 2pi/{n} not aligned with a {n}-fold rotational axis of the Octahedron.')

        return ans

    else:
        raise ValueError('The rotation order must be one of {2, 3, 4}.')


#############################################
# SUBGROUPS MAPS
#############################################

# C_N #####################################

def octa_to_cn(adj: GroupElement, cn: escnn.group.CyclicGroup, axis: np.ndarray):
    assert isinstance(adj.group, Octahedral)

    assert axis.shape == (3,)
    assert np.isclose(np.linalg.norm(axis), 1.)

    assert cn.order() in [2, 3, 4]

    assert _is_axis_aligned(axis, cn.order())

    def _map(e: GroupElement, cn=cn, adj=adj, axis=axis):
        octa = adj.group
        assert e.group == octa

        e = adj @ e @ (~adj)

        e = e.to('Q')

        v = e[:3]

        n = np.linalg.norm(v)

        if np.allclose(n, 0.):
            return cn.identity
        elif np.allclose(v / n, axis):
            # if the rotation is along the axis
            s, c = n, e[-1]
            theta = 2 * np.arctan2(s, c)
            try:
                return cn.element(theta, 'radians')
            except ValueError:
                return None
        else:
            return None

    return _map


def cn_to_octa(adj: GroupElement, cn: escnn.group.CyclicGroup, axis: np.ndarray):
    assert isinstance(adj.group, Octahedral)

    assert axis.shape == (3,)
    assert np.isclose(np.linalg.norm(axis), 1.)

    assert cn.order() in [2, 3, 4]

    assert _is_axis_aligned(axis, cn.order())

    def _map(e: GroupElement, cn=cn, adj=adj, axis=axis):
        assert e.group == cn
        octa = adj.group

        theta_2 = e.to('radians') / 2.

        q = np.empty(4)
        q[:3] = axis * np.sin(theta_2)
        q[-1] = np.cos(theta_2)

        return (~adj) @ octa.element(q, 'Q') @ adj

    return _map


#############################################
# Generate irreps
#############################################

from joblib import Memory
from escnn.group import __cache_path__

cache = Memory(__cache_path__, verbose=2)


@cache.cache(ignore=['octa'])
def _build_octa_irrep(octa: Octahedral, l: int):
    
    if l == -1:
        
        # matrix coefficients from https://arxiv.org/pdf/1110.6376.pdf
        
        # the matrix coefficients there are expressed wrt a different set of generators
        # we fist build this set of generators
        
        r3 = octa.generators[0]
        r = r3 @ r3 @ r3
        
        k = octa.elements[0]
        s = octa.generators[1]
        t = ~s @ k @ s @ r
        
        # Representation of `t`
        rho_t = np.array([
            [1., 0., 0.],
            [0., 0., 1.],
            [0., 1., 0.],
        ])
        
        # Representation of `k`
        rho_k = np.array([
            [1., 0., 0.],
            [0., -1., 0.],
            [0., 0., -1.],
        ])
        
        # Representation of `s`
        rho_s = np.array([
            [0., 1., 0.],
            [0., 0., 1.],
            [1., 0., 0.],
        ])
        
        #  https://arxiv.org/pdf/1110.6376.pdf defines the irrep `l = 1` (denoted by 3 there) as our
        #  `standard_representation`, which is expressed on a different basis than the Wigner D matrix with l=1.
        # Since `l=-1` (their 3') is defined as the tensor product between `l=1` and `l=3` (their 1')
        # we apply the inverse change of basis used in `standard_representation` to ensure that
        # `-1 = 1 \tensor 3` for us as well
        
        change_of_basis = np.array([
            [0, 0, 1],
            [1, 0, 0],
            [0, 1, 0]
        ])
        
        rho_t = change_of_basis.T @ rho_t @ change_of_basis
        rho_k = change_of_basis.T @ rho_k @ change_of_basis
        rho_s = change_of_basis.T @ rho_s @ change_of_basis
        
        generators = [
            (t, rho_t),
            (k, rho_k),
            (s, rho_s),
        ]
        
        return generate_irrep_matrices_from_generators(octa, generators)

    elif l == 2:

        # matrix coefficients from https://arxiv.org/pdf/1110.6376.pdf

        # the matrix coefficients there are expressed wrt a different set of generators
        # we fist build this set of generators

        r3 = octa.generators[0]
        r = r3 @ r3 @ r3

        k = octa.elements[0]
        s = octa.generators[1]
        t = ~s @ k @ s @ r

        # Representation of `t`
        rho_t = np.array([
            [0., 1.],
            [1., 0.],
        ])

        # Representation of `k`
        rho_k = np.array([
            [1., 0.],
            [0., 1.],
        ])

        # Representation of `s`
        rho_s = 0.5 * np.array([
            [-1., -np.sqrt(3)],
            [np.sqrt(3), -1.],
        ])

        generators = [
            (t, rho_t),
            (k, rho_k),
            (s, rho_s),
        ]
        
        return generate_irrep_matrices_from_generators(octa, generators)

    elif l == 3:

        # matrix coefficients from https://arxiv.org/pdf/1110.6376.pdf

        # the matrix coefficients there are expressed wrt a different set of generators
        # we fist build this set of generators

        r3 = octa.generators[0]
        r = r3 @ r3 @ r3

        k = octa.elements[0]
        s = octa.generators[1]
        t = ~s @ k @ s @ r

        # Representation of `t`
        rho_t = np.array([[-1.]])

        # Representation of `k`
        rho_k = np.array([[1.]])

        # Representation of `s`
        rho_s = np.array([[1.]])

        generators = [
            (t, rho_t),
            (k, rho_k),
            (s, rho_s),
        ]

        return generate_irrep_matrices_from_generators(octa, generators)

    else:
        raise ValueError()



from escnn.nn import GeometricTensor
from escnn.nn import FieldType
from .equivariant_module import EquivariantModule
import torch

from typing import List, Tuple, Union, Any

__all__ = ["IdentityModule"]


class IdentityModule(EquivariantModule):
    def __init__(self,
                 in_type: FieldType
                 ):
        r"""

        Simple module which does not perform any operation on the input tensor.

        Args:
            in_type (FieldType): input (and output) type of this module

        """
    
        super(IdentityModule, self).__init__()

        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type
        
    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        
        Returns the input tensor.
        
        Args:
            input (GeometricTensor): the input GeometricTensor

        Returns:
            the output tensor
            
        """
        
        assert input.type == self.in_type
        return input

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        
        assert len(input_shape) > 1
        assert input_shape[1] == self.in_type.size
        
        return input_shape

    def check_equivariance(self, atol: float = 2e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
        return super(IdentityModule, self).check_equivariance(atol=atol, rtol=rtol)

    def export(self):
        r"""
        Export this module to a normal PyTorch :class:`torch.nn.Identity` module and set to "eval" mode.
        
        .. warning ::
            Only working with PyTorch >= 1.2

        """
        self.eval()
        return torch.nn.Identity()




import numpy as np

from scipy.ndimage import rotate
from scipy.ndimage import affine_transform


def rotate_array_2d(x, angle):
    k = 2 * angle / np.pi
    if k.is_integer():
        # Rotations by 180 and 270 degrees seem to be not perfect using `ndimage.rotate` and can therefore
        # make some tests fail.
        # For this reason, we use `np.rot90` to perform rotations by multiples of 90 degrees without interpolation
        return np.rot90(x, k, axes=(-2, -1))
    else:
        return rotate(x, angle * 180.0 / np.pi, (-2, -1), reshape=False, order=2)


def linear_transform_array_3d(x, trafo: np.ndarray, exact=True, order=2):
    assert trafo.shape == (3, 3)
    assert len(x.shape) > 2
    return linear_transform_array_nd(x, trafo, exact, order=order)
    

def linear_transform_array_nd(x, trafo: np.ndarray, exact=True, order=2):
    n = trafo.shape[0]
    assert trafo.shape == (n, n)
    assert len(x.shape) >= n

    # TODO : MAKE THIS EXPLICIT SOMEWHERE IN THE DOCS!!!
    # assume trafo matrix has [X, Y, Z, ....] order
    # but input tensor has [..., -Z, -Y, X] order
    trafo = trafo[::-1, ::-1].copy()
    trafo[:-1, :] *= -1
    trafo[:, :-1] *= -1
    
    D = len(x.shape)
    at = np.abs(trafo)
    
    if exact and (
            np.isclose(at.sum(axis=0), 1).all() and
            np.isclose(at.sum(axis=1), 1).all() and
            (np.isclose(at, 1.) | np.isclose(at, 0.)).all()
    ):
        # if it is a permutation matrix we can perform this transformation without interpolation
        axs = np.around(trafo).astype(np.int) @ np.arange(1, n+1).reshape(n, 1)
        axs = axs.reshape(-1)
        
        stride = np.sign(axs).tolist()
        axs = np.abs(axs).tolist()
        
        axs = list(range(D - n)) + [D - n - 1 + a for a in axs]
        assert len(axs) == D, (len(axs), D)

        y = x.transpose(axs)
        
        stride = (Ellipsis,) + tuple([slice(None, None, s) for s in stride])
        y = y[stride]
        return y
    else:
        
        trafo = trafo.T

        t = np.eye(D)
        t[-n:, -n:] = trafo
        center = np.zeros(len(x.shape))
        center[-n:] = (np.asarray(x.shape[-n:]) - 1) / 2
        center[-n:] = -(trafo - np.eye(n)) @ center[-n:]

        return affine_transform(x, t, offset=center, order=order)


if __name__ == '__main__':

    # test that the exact rotation method produces the same results as the interpolation one
    # on all 48 origin-preserving isometries of the voxel grid
    
    import itertools
    x = np.random.randn(15, 15, 15)
    for p in itertools.permutations([0,1,2]):
        M = np.eye(3)[p, :]
        
        for s in itertools.product([-1, 1], repeat=3):
            rot = np.asarray(s).reshape(-1, 1) * M
            
            y1 = linear_transform_array_3d(x, rot, True)
            y2 = linear_transform_array_3d(x, rot, False, order=2)
            y3 = linear_transform_array_3d(x, rot, False, order=3)
            assert np.allclose(y2, y1), rot
            assert np.allclose(y3, y1), rot
            
    # test that the nd method is equivalent to the 2d one
    x = np.random.randn(3, 2, 11, 11)
    np.set_printoptions(suppress=True, precision=3, linewidth=100000)

    for _ in range(10):
        angle = np.random.rand()*2*np.pi
        
        rot = np.asarray([
            [np.cos(angle), -np.sin(angle)],
            [np.sin(angle), np.cos(angle)],
        ])
    
        y1 = rotate_array_2d(x, angle)
        y2 = linear_transform_array_nd(x, rot)
        assert np.allclose(y2, y1), rot




__all__ = [
    "__title__",
    "__summary__",
    "__url__",
    "__version__",
    "__author__",
    "__email__",
    "__license__",
]

__title__ = "escnn"
__summary__ = "E(n)-Equivariant CNNs Library for PyTorch"
__url__ = 'https://github.com/QUVA-Lab/escnn'
__version__ = "1.0.0"
__author__ = "Gabriele Cesa"
__email__ = "cesa.gabriele@gmail.com"
__license__ = "BSD 3-Clause Clear"




from abc import ABC, abstractmethod

from typing import List, Iterable, Dict, Union


__all__ = ["BasisManager"]


class BasisManager(ABC):
    
    def __init__(self):
        r"""
        Abstract class defining the *interface* for the different modules which deal with the filter basis.
        It provides a few methods which can be used to retrieve information about the basis and each of its elements.
       
        """
        super(BasisManager, self).__init__()
        
    @abstractmethod
    def get_element_info(self, name: int) -> Dict:
        """
        Method that returns the information associated to a basis element
        
        Parameters:
            name (int): index of the basis element
        
        Returns:
            dictionary containing the information
        """
        pass
    
    @abstractmethod
    def get_basis_info(self) -> Iterable[Dict]:
        """
        Method that returns an iterable over all basis elements' attributes.

        Returns:
            an iterable over all the basis elements' attributes
            
        """
        pass

    @abstractmethod
    def dimension(self) -> int:
        r"""
        The dimensionality of the basis and, so, the number of weights needed to expand it.
        
        Returns:
            the dimensionality of the basis
            
        """
        pass






from setuptools import setup, find_packages

about = {}
with open("escnn/__about__.py") as fp:
    exec(fp.read(), about)

install_requires = [
    'torch',
    'numpy',
    'scipy',
    'lie_learn',
    'joblib',
    'pymanopt',
    'autograd',
    'py3nj'
]


setup_requires = []
tests_require = ['scikit-learn', 'scikit-image']

with open("README.md", 'r', encoding='utf-8') as f:
    long_description = f.read()

download_url = 'https://github.com/QUVA-Lab/escnn/archive/v{}.tar.gz'.format(about['__version__'])

setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__summary__'],
    author=about['__author__'],
    author_email=about['__email__'],
    url=about['__url__'],
    download_url=download_url,
    license=about['__license__'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=['test', 'test.*']),
    python_requires='>=3.7',
    keywords=[
        'pytorch',
        'cnn',
        'convolutional-networks'
        'equivariant',
        'isometries',
    ],
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
)




from escnn.group import *
from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule

import torch
import torch.nn.functional as F

from typing import List, Tuple, Any

import numpy as np

__all__ = ["FourierPointwise", "FourierELU"]


def _build_kernel(G: Group, irrep: List[tuple]):
    kernel = []
    
    for irr in irrep:
        irr = G.irrep(*irr)
        
        c = int(irr.size//irr.sum_of_squares_constituents)
        k = irr(G.identity)[:, :c] * np.sqrt(irr.size)
        kernel.append(k.T.reshape(-1))
    
    kernel = np.concatenate(kernel)
    return kernel
    

class FourierPointwise(EquivariantModule):
    
    def __init__(self, gspace: GSpace, channels: int, irreps: List, *grid_args,  function: str = 'p_relu', inplace: bool = True, **grid_kwargs):
        r"""
        
        Applies a Inverse Fourier Transform to sample the input features, apply the pointwise non-linearity in the
        group domain (Dirac-delta basis) and, finally, computes the Fourier Transform to obtain irreps coefficients.
        
        .. note::
            This operation is only *approximately* equivariant and its equivariance depends on the sampling grid and the
            non-linear activation used, as well as the original band-limitation of the input features.
            
        The same function is applied to every channel independently.
        The input representation is preserved by this operation and, therefore, it equals the output representation.
        
        The class first constructs a band-limited regular representation of ```gspace.fibergroup``` using
        :meth:`escnn.group.Group.spectral_regular_representation`.
        The band-limitation of this representation is specified by ```irreps``` which should be a list containing
        a list of ids identifying irreps of ```gspace.fibergroup``` (see :attr:`escnn.group.IrreducibleRepresentation.id`).
        This representation is used to define the input and output field types, each containing ```channels``` copies
        of a feature field transforming according to this representation.
        A feature vector transforming according to such representation is interpreted as a vector of coefficients
        parameterizing a function over the group using a band-limited Fourier basis.
        
        To approximate the Fourier transform, this module uses a finite number of samples from the group.
        The set of samples used is specified by the ```grid_args``` and ```grid_kwargs``` which are passed to the method
        :meth:`~escnn.group.Group.grid`.
        
        .. todo::
            Mention the normalization of the transform we use
        
        Args:
            gspace (GSpace):  the gspace describing the symmetries of the data. The Fourier transform is
                              performed over the group ```gspace.fibergroup```
            channels (int): number of independent fields in the input `FieldType`
            irreps (list): list of irreps' ids to construct the band-limited representation
            function (str): the identifier of the non-linearity.
                    It is used to specify which function to apply.
                    By default (``'p_relu'``), ReLU is used.
            inplace (bool): applies the non-linear activation in-place. Default: `True`
            *grid_args: parameters used to construct the discretization grid
            **grid_kwargs: keyword parameters used to construct the discretization grid
            
        """

        assert isinstance(gspace, GSpace)
        
        super(FourierPointwise, self).__init__()

        self.space = gspace
        
        G: Group = gspace.fibergroup
        
        self.rho = G.spectral_regular_representation(*irreps, name=None)

        # the representation in input is preserved
        self.in_type = self.out_type = FieldType(self.space, [self.rho]*channels)
        
        # retrieve the activation function to apply
        if function == 'p_relu':
            self._function = F.relu_ if inplace else F.relu
        elif function == 'p_elu':
            self._function = F.elu_ if inplace else F.elu
        elif function == 'p_sigmoid':
            self._function = torch.sigmoid_ if inplace else F.sigmoid
        elif function == 'p_tanh':
            self._function = torch.tanh_ if inplace else F.tanh
        else:
            raise ValueError('Function "{}" not recognized!'.format(function))
        
        kernel = _build_kernel(G, irreps)
        assert kernel.shape[0] == self.rho.size
        
        kernel = kernel / np.linalg.norm(kernel)
        kernel = kernel.reshape(-1, 1)
        
        grid = G.grid(*grid_args, **grid_kwargs)
        
        A = np.concatenate(
            [
                self.rho(g) @ kernel
                for g in grid
            ], axis=1
        ).T
        
        eps = 1e-8
        Ainv = np.linalg.inv(A.T @ A + eps * np.eye(self.rho.size)) @ A.T
        
        self.register_buffer('A', torch.tensor(A, dtype=torch.get_default_dtype()))
        self.register_buffer('Ainv', torch.tensor(Ainv, dtype=torch.get_default_dtype()))
        
    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""

        Applies the pointwise activation function on the input fields

        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map after the non-linearities have been applied

        """

        assert input.type == self.in_type
        
        shape = input.shape
        x_hat = input.tensor.view(shape[0], len(self.in_type), self.rho.size, *shape[2:])
        
        x = torch.einsum('bcf...,gf->bcg...', x_hat, self.A)
        
        y = self._function(x)

        y_hat = torch.einsum('bcg...,fg->bcf...', y, self.Ainv)
        
        y_hat = y_hat.reshape(*shape)

        return GeometricTensor(y_hat, self.out_type, input.coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:

        assert len(input_shape) >= 2
        assert input_shape[1] == self.in_type.size

        b, c = input_shape[:2]
        spatial_shape = input_shape[2:]

        return (b, self.out_type.size, *spatial_shape)

    def check_equivariance(self, atol: float = 1e-4, rtol: float = 2e-2) -> List[Tuple[Any, float]]:
    
        c = self.in_type.size
        B = 128
        x = torch.randn(B, c, *[3]*self.space.dimensionality)
    
        errors = []
    
        # for el in self.space.testing_elements:
        for _ in range(100):
            
            el = self.space.fibergroup.sample()
    
            x1 = GeometricTensor(x.clone(), self.in_type)
            x2 = GeometricTensor(x.clone(), self.in_type).transform_fibers(el)

            out1 = self(x1).transform_fibers(el)
            out2 = self(x2)
            
            out1 = out1.tensor.view(B, len(self.in_type), self.rho.size, *out1.shape[2:]).detach().numpy()
            out2 = out2.tensor.view(B, len(self.in_type), self.rho.size, *out2.shape[2:]).detach().numpy()

            errs = np.linalg.norm(out1 - out2, axis=2).reshape(-1)
            errs[errs < atol] = 0.
            norm = np.sqrt(np.linalg.norm(out1, axis=2).reshape(-1) * np.linalg.norm(out2, axis=2).reshape(-1))
            
            relerr = errs / norm
            
            # print(el, errs.max(), errs.mean(), relerr.max(), relerr.min())
        
            assert relerr.mean()+ relerr.std() < rtol, \
                'The error found during equivariance check with element "{}" is too high: max = {}, mean = {}, std ={}' \
                    .format(el, relerr.max(), relerr.mean(), relerr.std())
        
            errors.append((el, errs.mean()))
    
        return errors


class FourierELU(FourierPointwise):
    
    def __init__(self, gspace: GSpace, channels: int, irreps: List, *grid_args, inplace: bool = True, **grid_kwargs):
        r"""

        Applies a Inverse Fourier Transform to sample the input features, apply ELU point-wise in the
        group domain (Dirac-delta basis) and, finally, computes the Fourier Transform to obtain irreps coefficients.
        See :class:`~escnn.nn.FourierPointwise` for more details.

        Args:
            gspace (GSpace):  the gspace describing the symmetries of the data. The Fourier transform is
                              performed over the group ```gspace.fibergroup```
            channels (int): number of independent fields in the input `FieldType`
            irreps (list): list of irreps' ids to construct the band-limited representation
            inplace (bool): applies the non-linear activation in-place. Default: `True`
            *grid_args: parameters used to construct the discretization grid
            **grid_kwargs: keyword parameters used to construct the discretization grid

        """
        
        super(FourierELU, self).__init__(gspace, channels, irreps, *grid_args, function='p_elu', inplace=inplace, **grid_kwargs)





from escnn.gspaces import *
from escnn.group import CyclicGroup
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule

import torch

from typing import List, Tuple, Any

import numpy as np

__all__ = ["VectorFieldNonLinearity"]


class VectorFieldNonLinearity(EquivariantModule):
    
    def __init__(self, in_type: FieldType, **kwargs):
        r"""
        
        VectorField non-linearities.
        This non-linearity only supports the regular representation of cyclic group :math:`C_N`, i.e. the group of
        :math:`N` discrete rotations.
        For each input field, the output one is built by taking the rotation associated with the highest
        activation; then, a 2-dimensional vector with an angle with respect to the x-axis equal to that rotation and a
        length equal to its activation is set in the output field.
        
        Args:
            in_type (FieldType): the input field type
            
        """
        assert isinstance(in_type.gspace, GSpace)

        assert isinstance(in_type.gspace.fibergroup, CyclicGroup)
        assert in_type.gspace.fibergroup.order() > 1
        
        for r in in_type.representations:
            assert 'vectorfield' in r.supported_nonlinearities,\
                'Error! Representation "{}" does not support "vector-field" non-linearity'.format(r.name)
            
            assert r.name == 'regular' and r.size == in_type.gspace.fibergroup.order(), r.name

        super(VectorFieldNonLinearity, self).__init__()

        self.space = in_type.gspace
        self.in_type = in_type
        
        # build the output representation substituting each input field with a rotation representation with frequency 1
        self.out_type = FieldType(self.space, [self.space.representations['irrep_1']] * len(in_type))
        
        # the number of rotations associated with the group action
        self._rotations = self.space.fibergroup.order()
        
    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        
        Apply the VectorField non-linearity to the input feature map.
        
        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map
            
        """
        
        assert input.type == self.in_type

        b, c = input.shape[:2]
        spatial_shape = input.shape[2:]

        # split the channel dimension in 2 dimensions, separating fields
        fm = input.tensor.view(b, -1, self._rotations, *spatial_shape)
        
        # evaluate the base rotation associated with the group action
        base_angle = 2 * np.pi / self._rotations
        
        # for each field, retrieve the maximum activation (and the argmax)
        max_activations, argmaxes = torch.max(fm, 2)
        max_activations = torch.relu_(max_activations)
        
        # compute the angles from the index of the maximum activation in the field
        max_angles = argmaxes.to(dtype=torch.float) * base_angle
        
        # build the output tensor
        output = torch.empty(b, self.out_type.size, *spatial_shape, dtype=torch.float, device=input.tensor.device)
        
        # to build the output vectors, take the cosine and the sine of the argmax angle
        # and multiply the 2-dimensional vector by the activation value
        output[:, ::2, ...] = torch.cos(max_angles) * max_activations
        output[:, 1::2, ...] = torch.sin(max_angles) * max_activations
        
        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, input.coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:

        assert len(input_shape) >= 2
        assert input_shape[1] == self.in_type.size

        b, c = input_shape[:2]
        spatial_shape = input_shape[2:]

        return (b, self.out_type.size, *spatial_shape)

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
    
        c = self.in_type.size
    
        x = torch.randn(3, c, 10, 10)
    
        x = GeometricTensor(x, self.in_type)
    
        errors = []
    
        for el in self.space.testing_elements:
            out1 = self(x).transform_fibers(el)
            out2 = self(x.transform_fibers(el))
        
            errs = (out1.tensor - out2.tensor).detach().numpy()
            errs = np.abs(errs).reshape(-1)
            print(el, errs.max(), errs.mean(), errs.var())
        
            assert torch.allclose(out1.tensor, out2.tensor, atol=atol, rtol=rtol), \
                'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}' \
                    .format(el, errs.max(), errs.mean(), errs.var())
        
            errors.append((el, errs.mean()))
    
        return errors





from __future__ import annotations

from escnn import gspaces
from escnn import kernels

from escnn.group import *

from .utils import linear_transform_array_3d

import numpy as np

from typing import Tuple, Union, Callable, List

__all__ = [
    "GSpace3D",
    ###########
    "flipRot3dOnR3",
    "rot3dOnR3",
    "fullIcoOnR3",
    "icoOnR3",
    "fullOctaOnR3",
    "octaOnR3",
    "dihedralOnR3",
    "rot2dOnR3",
    "conicalOnR3",
    "fullCylindricalOnR3",
    "cylindricalOnR3",
    "mirOnR3",
    "invOnR3",
    "trivialOnR3"
]


class GSpace3D(gspaces.GSpace):
    
    def __init__(self, sg_id: Tuple, maximum_frequency: int = 2):
        r"""

        Describes reflectional and rotational symmetries of the plane :math:`\R^3`.
        
        .. note ::
            A point :math:`\bold{v} \in \R^3` is parametrized using an :math:`(X, Y, Z)` convention,
            i.e. :math:`\bold{v} = (x, y, z)^T`.
            The representation :attr:`escnn.gspaces.GSpace3D.basespace_action` also assumes this convention.
            
            However, when working with voxel data, the :math:`(-Z, -Y, X)` convention is used.
            That means that, in a 5-dimensional feature tensor of shape ``(B, C, D1, D2, D3)``, the last dimension
            is the X axis, the second last the (inverted) Y axis and then the (inverted) Z axis.
            Note that this is consistent with 2D images, where a :math:`(-Y, X)` convention is used.
            
            This is especially relevant when transforming a :class:`~escnn.nn.GeometricTensor` or when building
            convolutional filters in :class:`~escnn.nn.R3Conv` which should be equivariant to subgroups of :math:`\O3`
            (e.g. when choosing the rotation axis for :func:`~escnn.gspaces.rot2dOnR3`).

        """

        o3 = o3_group(maximum_frequency=maximum_frequency)
        _sg_id = o3._process_subgroup_id(sg_id)
        fibergroup, inclusion, restriction = o3.subgroup(_sg_id)
        
        # TODO - catch sg_id and build a dictionary of more meaningful names
        # use the input sg_id instead of the processed one to avoid adding the adjoint parameter unless specified
        name = f'{fibergroup}_on_R3[{sg_id}]'

        self._sg_id = _sg_id
        self._inclusion = inclusion
        self._restriction = restriction
        self._base_action = o3.standard_representation().restrict(_sg_id)
        
        super(GSpace3D, self).__init__(fibergroup, 3, name)
    
    def restrict(self, id: Tuple) -> Tuple[gspaces.GSpace, Callable, Callable]:
        r"""

        Build the :class:`~escnn.group.GSpace` associated with the subgroup of the current fiber group identified by
        the input ``id``

        Args:
            id (tuple): the id of the subgroup

        Returns:
            a tuple containing

                - **gspace**: the restricted gspace

                - **back_map**: a function mapping an element of the subgroup to itself in the fiber group of the original space

                - **subgroup_map**: a function mapping an element of the fiber group of the original space to itself in the subgroup (returns ``None`` if the element is not in the subgroup)

        """

        o3 = o3_group()
        sg_id = o3._combine_subgroups(self._sg_id, id)
        sg, inclusion, restriction = self.fibergroup.subgroup(id)

        return GSpace3D(sg_id), inclusion, restriction

    def _basis_generator(self,
                         in_repr: Representation,
                         out_repr: Representation,
                         rings: List[float],
                         sigma: List[float],
                         **kwargs,
                         ) -> kernels.KernelBasis:
        r"""
        Method that builds the analitical basis that spans the space of equivariant filters which
        are intertwiners between the representations induced from the representation ``in_repr`` and ``out_repr``.

        `kwargs` can be used to specify `maximum_frequency`
        
        Args:
            in_repr (Representation): the input representation
            out_repr (Representation): the output representation
            rings (list): radii of the rings where to sample the bases
            sigma (list): parameters controlling the width of each ring where the bases are sampled.

        Returns:
            the basis built

        """
        
        # TODO - add max_offset for cyclic and dihedral groups!

        if 'maximum_frequency' in kwargs:
            maximum_frequency = kwargs['maximum_frequency']
        else:
            maximum_frequency = None

        if self._sg_id == (True, 'so3'):
            return kernels.kernels_O3_act_R3(in_repr, out_repr, rings, sigma, maximum_frequency=maximum_frequency, adjoint=None)
        elif self._sg_id == (False, 'so3'):
            return kernels.kernels_SO3_act_R3(in_repr, out_repr, rings, sigma, maximum_frequency=maximum_frequency, adjoint=None)
        elif self._sg_id[0] == False:
            sg_id = self._sg_id[1:]
            if isinstance(sg_id[-1], GroupElement):
                # the adjoint is an O(3) group element
                # convert it to an SO(3) element
                # not that even if the adjoint contains the 3d inversion, we can ignore it
                # (since O(3) is a direct product, the inversion commutes with any 3D rotation)
                adj = sg_id[-1]
                so3 = so3_group()
                adj = so3.element(adj.value[1], adj.param)
                sg_id = sg_id[:-1] + (adj,)
            return kernels.kernels_SO3_subgroup_act_R3(in_repr, out_repr, sg_id, rings, sigma, maximum_frequency=maximum_frequency, adjoint=None)
        else:
            return kernels.kernels_O3_subgroup_act_R3(in_repr, out_repr, self._sg_id, rings, sigma, maximum_frequency=maximum_frequency, adjoint=None)

    @property
    def basespace_action(self) -> Representation:
        return self._base_action

    def __eq__(self, other):
        if isinstance(other, GSpace3D):
            return self._sg_id == other._sg_id
        else:
            return False
    
    def __hash__(self):
        return 1000 * hash(self.name) + hash(self._sg_id)


########################################################################################################################


def flipRot3dOnR3(maximum_frequency: int = 2) -> GSpace3D:
    r"""
    Describes 3D rotation and inversion symmetries in the space :math:`\R^3`.
    
    .. todo ::
        rename to invRot3dOnR3?

    Args:
        maximum_frequency (int): maximum frequency of :class:`~e2cnn.group.O3`'s irreps

    """
    sg_id = 'o3'
    return GSpace3D(sg_id, maximum_frequency=maximum_frequency)
    
    
def rot3dOnR3(maximum_frequency: int = 2) -> GSpace3D:
    r"""
    Describes 3D rotation symmetries in the space :math:`\R^3`.
    
    Args:
        maximum_frequency (int): maximum frequency of :class:`~e2cnn.group.SO3`'s irreps
    
    """
    sg_id = 'so3'
    return GSpace3D(sg_id, maximum_frequency=maximum_frequency)


def fullIcoOnR3() -> GSpace3D:
    sg_id = True, 'ico'
    return GSpace3D(sg_id, maximum_frequency=4)


def icoOnR3() -> GSpace3D:
    r"""
    Describes 3D rotation symmetries of a Icosahedron (or Dodecahedron) in the space :math:`\R^3`
    """
    sg_id = False, 'ico'
    return GSpace3D(sg_id, maximum_frequency=4)


def fullOctaOnR3() -> GSpace3D:
    sg_id = True, 'octa'
    return GSpace3D(sg_id, maximum_frequency=3)


def octaOnR3() -> GSpace3D:
    r"""
    Describes 3D rotation symmetries of an Octahedron (or Cube) in the space :math:`\R^3`
    """
    sg_id = False, 'octa'
    return GSpace3D(sg_id, maximum_frequency=3)


def dihedralOnR3(n: int = -1, axis: float = np.pi / 2, adjoint: GroupElement = None, maximum_frequency: int = 2) -> GSpace3D:
    r"""
    Describes 2D rotation symmetries along the :math:`Z` axis in the space :math:`\R^3` and :math:`\pi` rotations
    along the ``axis`` in the :math:`XY` plane, i.e. the rotations inside the plane :math:`XY` and reflections around
    the ``axis``.

    The ``adjoint`` parameter can be a :class:`~escnn.group.GroupElement` of :class:`~escnn.group.O3`.
    If not ``None`` (which is equivalent to the identity), this specifies another :math:`\SO2` subgroup of :math:`\O3`
    which is adjoint to the :math:`\SO2` subgroup of rotations around the :math:`Z` axis.
    If ``adjoint`` is the group element :math:`A \in \O3`, the new subgroup would then represent rotations around the
    axis :math:`A^{-1} \cdot (0, 0, 1)^T`.

    If ``N > 1``, the gspace models *discrete* rotations by angles which are multiple of :math:`\frac{2\pi}{N}`
    (:class:`~e2cnn.group.CyclicGroup`).
    Otherwise, if ``N=-1``, the gspace models *continuous* planar rotations (:class:`~e2cnn.group.SO2`).
    In that case the parameter ``maximum_frequency`` is required to specify the maximum frequency of the irreps of
    :class:`~e2cnn.group.SO2` (see its documentation for more details)

    Args:
        N (int): number of discrete rotations (integer greater than 1) or ``-1`` for continuous rotations
        adjoint (GroupElement, optional): an element of :math:`\O3`
        maximum_frequency (int): maximum frequency of :class:`~e2cnn.group.SO2`'s irreps if ``N = -1``

    """
    assert isinstance(n, int)
    assert n == -1 or n > 0
    
    sg_id = False, 2*axis, n
    
    if adjoint is not None:
        sg_id += (adjoint,)
    
    return GSpace3D(sg_id, maximum_frequency=maximum_frequency)
    
    
def rot2dOnR3(n: int = -1, adjoint: GroupElement = None, maximum_frequency: int = 2) -> GSpace3D:
    r"""

    Describes 2D rotation symmetries along the :math:`Z` axis in the space :math:`\R^3`, i.e. the rotations inside the
    plane :math:`XY`.
    
    ``adjoint`` is a :class:`~escnn.group.GroupElement` of :class:`~escnn.group.O3`.
    If not ``None`` (which is equivalent to the identity), this specifies another :math:`\SO2` subgroup of :math:`\O3`
    which is adjoint to the :math:`\SO2` subgroup of rotations around the :math:`Z` axis.
    If ``adjoint`` is the group element :math:`A \in \O3`, the new subgroup would then represent rotations around the
    axis :math:`A^{-1} \cdot (0, 0, 1)^T`.

    If ``N > 1``, the gspace models *discrete* rotations by angles which are multiple of :math:`\frac{2\pi}{N}`
    (:class:`~e2cnn.group.CyclicGroup`).
    Otherwise, if ``N=-1``, the gspace models *continuous* planar rotations (:class:`~e2cnn.group.SO2`).
    In that case the parameter ``maximum_frequency`` is required to specify the maximum frequency of the irreps of
    :class:`~e2cnn.group.SO2` (see its documentation for more details)

    Args:
        N (int): number of discrete rotations (integer greater than 1) or ``-1`` for continuous rotations
        adjoint (GroupElement, optional): an element of :math:`\O3`
        maximum_frequency (int): maximum frequency of :class:`~e2cnn.group.SO2`'s irreps if ``N = -1``

    """
    assert isinstance(n, int)
    assert n == -1 or n > 0
    sg_id = False, False, n

    if adjoint is not None:
        sg_id += (adjoint,)
        
    return GSpace3D(sg_id, maximum_frequency=maximum_frequency)
    
    
def conicalOnR3(n: int = -1, axis: float = np.pi / 2., adjoint: GroupElement = None, maximum_frequency: int = 2) -> GSpace3D:
    assert isinstance(n, int)
    assert n == -1 or n > 0
    
    sg_id = 'cone', 2*axis, n
    
    if adjoint is not None:
        sg_id += (adjoint,)
        
    return GSpace3D(sg_id, maximum_frequency=maximum_frequency)
    
    
def mirOnR3(axis: float = np.pi / 2, adjoint: GroupElement = None) -> GSpace3D:
    r"""

    Describes mirroring with respect to a plane in the space :math:`\R^3`.
    
    .. todo ::
        Document what ``axis`` and ``adjoint`` describe or change parameters, just getting a :math:`\bold{v} \in \R^3`
        vector in input which specifies the mirroring axis.

    """

    sg_id = 'cone', 2*axis, 1
    
    if adjoint is not None:
        sg_id += (adjoint,)
    
    return GSpace3D(sg_id, maximum_frequency=1)


def fullCylindricalOnR3(n: int = -1, axis: float = np.pi / 2, adjoint: GroupElement = None, maximum_frequency: int = 2) -> GSpace3D:
    assert isinstance(n, int)
    assert n == -1 or n > 0
    
    sg_id = True, axis, n
    
    if adjoint is not None:
        sg_id += (adjoint,)
    
    return GSpace3D(sg_id, maximum_frequency=maximum_frequency)


def cylindricalOnR3(n: int = -1, adjoint: GroupElement = None, maximum_frequency: int = 2) -> GSpace3D:
    assert isinstance(n, int)
    assert n == -1 or n > 0
    
    sg_id = True, False, n
    
    if adjoint is not None:
        sg_id += (adjoint,)
    
    return GSpace3D(sg_id, maximum_frequency=maximum_frequency)


def invOnR3() -> GSpace3D:
    r"""

    Describes the inversion symmetry of the space :math:`\R^3`.

    An inversion flips the sign of all coordinates, mapping a vector :math:`\bold{v} \in \R^3` to :math:`-\bold{v}`.

    """
    sg_id = True, False, 1
    return GSpace3D(sg_id, maximum_frequency=1)
    
    
def trivialOnR3() -> GSpace3D:
    r"""
    Describes the space :math:`\R^3` without considering any origin-preserving symmetry.
    This is modeled by choosing trivial fiber group :math:`\{e\}`.

    .. note ::
        This models the symmetries of conventional *Convolutional Neural Networks* which are not equivariant to
        origin preserving transformations such as rotations and reflections.

    """
    sg_id = False, False, 1
    return GSpace3D(sg_id, maximum_frequency=1)

